# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import division  # Use Python3 divison in Python2

import os.path
import paramiko
import re

from nfv_tempest_plugin.tests.scenario import manager_utils
from oslo_log import log
from oslo_serialization import jsonutils
from tempest.api.compute import api_microversion_fixture
from tempest.common import waiters
from tempest import config
from tempest.lib.common import api_version_utils
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions as lib_exc
from tempest.scenario import manager
"""Python 2 and 3 support"""
from six.moves import StringIO


CONF = config.CONF
LOG = log.getLogger('{} [-] nfv_plugin_test'.format(__name__))


class BareMetalManager(api_version_utils.BaseMicroversionTest,
                       manager.ScenarioTest,
                       manager_utils.ManagerMixin):
    """This class Interacts with BareMetal settings"""
    credentials = ['primary', 'admin']

    def __init__(self, *args, **kwargs):
        super(BareMetalManager, self).__init__(*args, **kwargs)
        self.public_network = CONF.network.public_network_id
        self.mgmt_network = None
        self.instance_user = CONF.nfv_plugin_options.instance_user
        self.instance_pass = CONF.nfv_plugin_options.instance_pass
        self.nfv_scripts_path = CONF.nfv_plugin_options.transfer_files_dest
        self.flavor_ref = CONF.compute.flavor_ref
        self.test_all_provider_networks = \
            CONF.nfv_plugin_options.test_all_provider_networks
        self.cpuregex = re.compile('^[0-9]{1,2}$')
        self.external_config = None
        self.test_setup_dict = {}
        self.remote_ssh_sec_groups = []
        self.remote_ssh_sec_groups_names = []
        self.qos_policy_groups = []
        self.servers = []
        self.test_network_dict = {}
        self.test_flavor_dict = {}
        self.test_instance_repo = {}
        self.user_data = {}
        self.user_data_b64 = ''
        self.fip = True
        self.external_resources_data = None

    @classmethod
    def setup_clients(cls):
        super(BareMetalManager, cls).setup_clients()
        cls.hypervisor_client = cls.manager.hypervisor_client
        cls.aggregates_client = cls.manager.aggregates_client
        cls.volumes_client = cls.os_primary.volumes_client_latest
        """
        security groups client
        floating ip client to support
        nova microversion>=2.36 changes
        """
        cls.security_groups_client = (
            cls.os_primary.security_groups_client)
        cls.security_group_rules_client = (
            cls.os_primary.security_group_rules_client)
        cls.floating_ips_client = (
            cls.os_primary.floating_ips_client)

    def setUp(self):
        """Check hypervisor configuration:

        SSH user and Private key/password definition [must].
        External config file exist [not a must].
        """
        super(BareMetalManager, self).setUp()
        self.assertIsNotNone(CONF.nfv_plugin_options.overcloud_node_user,
                             "Missing SSH user login in config")

        if CONF.nfv_plugin_options.overcloud_node_pkey_file:
            key_str = open(
                CONF.nfv_plugin_options.overcloud_node_pkey_file).read()
            CONF.nfv_plugin_options.overcloud_node_pkey_file_rsa = \
                paramiko.RSAKey.from_private_key(StringIO(key_str))
        else:
            self.assertIsNotNone(
                CONF.nfv_plugin_options.overcloud_node_pass,
                'Missing SSH password or key_file')
        if CONF.nfv_plugin_options.external_config_file:
            if os.path.exists(CONF.nfv_plugin_options.external_config_file):
                self.read_external_config_file()

        self.useFixture(api_microversion_fixture.APIMicroversionFixture(
            self.request_microversion))

        if CONF.nfv_plugin_options.external_resources_output_file:
            if os.path.exists(
                    CONF.nfv_plugin_options.external_resources_output_file):
                self._read_and_validate_external_resources_data_file()

        if CONF.nfv_plugin_options.quota_cores and \
                CONF.nfv_plugin_options.quota_ram:
            self.os_admin.quotas_client.update_quota_set(
                self.os_primary.tenants_client.tenant_id,
                cores=CONF.nfv_plugin_options.quota_cores,
                ram=CONF.nfv_plugin_options.quota_ram)

    @classmethod
    def resource_setup(cls):
        super(BareMetalManager, cls).resource_setup()
        cls.tenant_id = cls.manager.identity_client.tenant_id
        cls.request_microversion = (
            api_version_utils.select_request_microversion(
                cls.min_microversion,
                CONF.compute.min_microversion))

    @classmethod
    def setup_credentials(cls):
        super(BareMetalManager, cls).setup_credentials()

    def create_flavor(self, name='flavor', ram='2048', disk='20', vcpus='1',
                      **flavor_args):
        """The method creates flavor based on the args passed to the method.

        The flavor could be created with or without an extra specs.
        In case method call with empty parameters, default values will
        be used and default flavor will be created.

        :param name: Flavor name.
        :param ram: Flavor ram.
        :param disk: Flavor disk.
        :param vcpus: Flavor vcpus.
        :param flavor_args: Dict of parameters for the flavor that should be
                created.
        :return flavor_id: ID of the created flavor.
        """
        flavor = self.os_admin.flavors_client.create_flavor(name=name,
                                                            ram=ram,
                                                            disk=disk,
                                                            vcpus=vcpus)
        flavor_id = flavor['flavor']['id']
        if 'extra_specs' in flavor_args:
            extra_specs = flavor_args['extra_specs']
            if isinstance(flavor_args['extra_specs'], list):
                extra_specs = flavor_args['extra_specs'][0]
            self.os_admin.flavors_client.set_flavor_extra_spec(flavor_id,
                                                               **extra_specs)
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self.os_admin.flavors_client.delete_flavor, flavor_id)
        return flavor_id

    def create_volume(self, **volume_args):
        """The method creates volume based on the args passed to the method.

        In case method call with empty parameters, default values will
        be used and default volume will be created.

        :param volume_args: Dict of parameters for the volume that should be
        created
        :return volume_id: ID of the created volume.
        """
        if 'name' not in volume_args:
            volume_args['name'] = data_utils.rand_name('volume')
        if 'size' not in volume_args:
            volume_args['size'] = CONF.volume.volume_size
        volume = self.volumes_client.create_volume(**volume_args)['volume']
        self.addClassResourceCleanup(
            self.volumes_client.wait_for_resource_deletion, volume['id'])
        self.addClassResourceCleanup(test_utils.call_and_ignore_notfound_exc,
                                     self.volumes_client.delete_volume,
                                     volume['id'])
        waiters.wait_for_volume_resource_status(self.volumes_client,
                                                volume['id'], 'available')
        return volume

    def _detach_volume(self, server, volume):
        """Detaches a volume and ignores if not found or not in-use

        param server: Created server details
        param volume: Created volume details
        """
        try:
            volume = self.volumes_client.show_volume(volume['id'])['volume']
            if volume['status'] == 'in-use':
                self.servers_client.detach_volume(server['id'], volume['id'])
        except lib_exc.NotFound:
            pass

    def attach_volume(self, server, volume):
        """Attaches volume to server

        param server: Created server details
        param volume: Created volume details
        :return volume_attachment: Volume attachment information.
        """
        attach_args = dict(volumeId=volume['id'])
        attachment = self.servers_client.attach_volume(
            server['id'], **attach_args)['volumeAttachment']
        self.addCleanup(waiters.wait_for_volume_resource_status,
                        self.volumes_client, volume['id'], 'available')
        self.addCleanup(self._detach_volume, server, volume)
        waiters.wait_for_volume_resource_status(self.volumes_client,
                                                volume['id'], 'in-use')
        return attachment

    def create_and_set_availability_zone(self, zone_name=None, **kwargs):
        """Create availability zone with aggregation

        The method creates an aggregate and add the availability zone label
        :param zone_name: Availability zone name
        :param kwargs:
                aggr_name: The name of the aggregation to be created
                hyper_hosts: The list of the hypervisors to be attached
                aggr_meta: The metadata for the aggregation
        """
        if not zone_name:
            zone_name = data_utils.rand_name('availability-zone')
        aggr = self.create_and_set_aggregate(**kwargs)
        zone = self.aggregates_client.update_aggregate(
            aggregate_id=aggr['id'], availability_zone=zone_name)
        return zone['aggregate']

    def create_and_set_aggregate(self, hyper_hosts, aggr_name=None,
                                 aggr_meta=None):
        """Create aggregation and add an hypervisor to it

        :param hyper_hosts: The list of the hypervisors to be attached
        :param aggr_name: The name of the aggregation to be created
        :param aggr_meta: The metadata for the aggregation (optional)
        """
        if not aggr_name:
            aggr_name = data_utils.rand_name('aggregate')
        hyper_list = []
        for hyper in self.hypervisor_client.list_hypervisors()['hypervisors']:
            for host in hyper_hosts:
                if hyper['hypervisor_hostname'].split('.')[0] in host:
                    hyper_list.append(hyper['hypervisor_hostname'])
        if not hyper_list:
            raise ValueError('Provided host for the aggregate does not exist.')

        aggr = self.aggregates_client.create_aggregate(name=aggr_name)
        self.addCleanup(self.aggregates_client.delete_aggregate,
                        aggr['aggregate']['id'])
        if aggr_meta:
            meta_body = {aggr_meta.split('=')[0]: aggr_meta.split('=')[1]}
            self.aggregates_client.set_metadata(aggr['aggregate']['id'],
                                                metadata=meta_body)
        for host in hyper_list:
            self.aggregates_client.add_host(aggr['aggregate']['id'], host=host)
            self.addCleanup(self.aggregates_client.remove_host,
                            aggr['aggregate']['id'], host=host)
        return aggr['aggregate']

    def _list_aggregate(self, name=None):
        """Aggregation listing

        This Method lists aggregation based on name, and returns the
        aggregated hosts lists.
        TBD: Add support to return, hosts list
        TBD: Return None in case no aggregation found.

        :param name
        """
        host = None

        if not name:
            return host

        aggregate = self.aggregates_client.list_aggregates()['aggregates']
        #       Assertion check
        if aggregate:
            aggr_result = []
            for i in aggregate:
                if name in i['name']:
                    aggr_result.append(self.aggregates_client.
                                       show_aggregate(i['id'])['aggregate'])
            host = aggr_result[0]['hosts']
        return host

    def _create_test_networks(self):
        """Method reads test-networks attributes from external_config.yml

        The network will be created for tempest tenant.
        Do not use this method if the test
        need to run on pre-configured networks..
        see _detect_existing_networks() method
        """
        if len(self.external_config['test-networks']) > 0:
            self.test_network_dict.clear()
        mgmt_network = None
        for net in self.external_config['test-networks']:
            self.test_network_dict[net['name']] = \
                {'provider:physical_network': net['physical_network'],
                 'provider:network_type': net['network_type'],
                 'dhcp': net['enable_dhcp'],
                 'cidr': net['cidr'],
                 'pool_start': net['allocation_pool_start'],
                 'pool_end': net['allocation_pool_end'],
                 'gateway_ip': net['gateway_ip'],
                 'port_type': net['port_type'],
                 'ip_version': net['ip_version']}
            if 'segmentation_id' in net:
                self.test_network_dict[net['name']][
                    'provider:segmentation_id'] = net['segmentation_id']
            if 'sec_groups' in net:
                self.test_network_dict[net['name']]['sec_groups'] = \
                    net['sec_groups']
            if 'mgmt' in net and net['mgmt']:
                mgmt_network = net['name']
            if 'mgmt' in net and 'dns_nameservers' in net:
                self.test_network_dict[net['name']]['dns_nameservers'] = \
                    net['dns_nameservers']
            if ('tag' in net and (2.32 <= float(self.request_microversion)
                                  <= 2.36 or float(self.request_microversion)
                                  >= 2.42)):
                self.test_network_dict[net['name']]['tag'] = net['tag']
            if 'trusted_vf' in net and net['trusted_vf']:
                self.test_network_dict[net['name']]['trusted_vf'] = True
            if 'switchdev' in net and net['switchdev']:
                self.test_network_dict[net['name']]['switchdev'] = True
            if 'min_qos' in net and net['min_qos']:
                self.test_network_dict[net['name']]['min_qos'] = \
                    net['min_qos']
            if net.get('skip_srv_attach') and net['skip_srv_attach']:
                self.test_network_dict[net['name']]['skip_srv_attach'] = True
        network_kwargs = {}
        """
        Create network and subnets
        """
        for net_name, net_param in iter(self.test_network_dict.items()):
            network_kwargs.clear()
            network_kwargs['name'] = net_name
            if 'sec_groups' in net_param and not net_param['sec_groups']:
                network_kwargs['port_security_enabled'] = net_param[
                    'sec_groups']
            """Added this for VxLAN no need of physical network or segmentation
            """
            if 'provider:network_type' in net_param and \
                    (net_param['provider:network_type'] == 'vlan'
                     or net_param['provider:network_type'] == 'flat'):
                if 'provider:physical_network' in net_param:
                    network_kwargs['provider:physical_network'] =\
                        net_param['provider:physical_network']
                if 'provider:segmentation_id' in net_param:
                    network_kwargs['provider:segmentation_id'] =\
                        net_param['provider:segmentation_id']

            if 'provider:network_type' in net_param:
                network_kwargs['provider:network_type'] =\
                    net_param['provider:network_type']

            network_kwargs['tenant_id'] = self.networks_client.tenant_id
            result = self.os_admin.networks_client.create_network(
                **network_kwargs)
            network = result['network']
            self.assertEqual(network['name'], net_name)
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            self.os_admin.networks_client.delete_network,
                            network['id'])
            network_kwargs.clear()
            network_kwargs['network_id'] = network['id']
            self.test_network_dict[net_name]['net-id'] = network['id']
            network_kwargs['name'] = net_name + '_subnet'
            network_kwargs['ip_version'] = net_param['ip_version']
            if 'cidr' in net_param:
                network_kwargs['cidr'] = net_param['cidr']
            if 'gateway_ip' in net_param:
                network_kwargs['gateway_ip'] = net_param['gateway_ip']
            if 'dhcp' in net_param:
                network_kwargs['enable_dhcp'] = net_param['dhcp']
            if 'pool_start' in net_param:
                network_kwargs['allocation_pools'] = \
                    [{'start': net_param['pool_start'],
                      'end':net_param['pool_end']}]
            if 'dns_nameservers' in net_param:
                network_kwargs['dns_nameservers'] = \
                    net_param['dns_nameservers']

            result = self.subnets_client.create_subnet(**network_kwargs)
            subnet = result['subnet']
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            self.subnets_client.delete_subnet, subnet['id'])
            self.test_network_dict[net_name]['subnet-id'] = subnet['id']
        if mgmt_network is not None:
            self.mgmt_network = mgmt_network

    def _add_subnet_to_router(self, mgmt_subnet_only=False):
        """Add subnets of existing networks to the router

        The subnets attached to the router will get the ability for the
        instances to get dhcp allocations on the instance interfaces

        :param mgmt_subnet_only: Attach only mgmt network subnet to router
        """
        mgmt_net_name = self.mgmt_network
        mgmt_net = self.test_network_dict[mgmt_net_name]

        subnets = []
        if not mgmt_subnet_only:
            nets = self.os_admin.networks_client.list_networks()['networks']
            for net in nets:
                if not net['router:external'] \
                        and 'HA network tenant' not in net['name']:
                    subnets.append(net['subnets'][0])
        else:
            subnets.append(mgmt_net['subnet-id'])
        seen_routers = self.os_admin.routers_client.list_routers()['routers']
        self.assertEqual(len(seen_routers), 1,
                         "Test require 1 admin router. please check")
        for subnet in subnets:
            self.os_admin.routers_client.add_router_interface(
                seen_routers[0]['id'], subnet_id=subnet)
            self.addCleanup(self.os_admin.routers_client.
                            remove_router_interface, seen_routers[0]['id'],
                            subnet_id=subnet)

    def _detect_existing_networks(self):
        """Use method only when test require no network

        cls.set_network_resources()
        it run over external_config networks,
        verified against existing networks..
        in case all networks exist return True and fill self.test_networks
        lists.
        In case there is external router.. public network decided
        based on router_external=False and router is not None
        """
        self.assertIsNotNone(CONF.nfv_plugin_options.external_config_file,
                             'This test require missing external_config, '
                             'for this test')

        self.assertTrue(self.test_network_dict,
                        'No networks for test, please check '
                        'external_config_file')

        public_network = self.networks_client.list_networks(
            **{'router:external': True})['networks']

        """
        Check public network exist in networks.
        remove it from network list
        if  = 0 we create port on first network if = 1  public network exist
        and set next network as vm management network
        name must not be public, router exist and network external false
        """
        if len(public_network) == 0:
            self.mgmt_network = self.test_network_dict.keys()[0]

        elif len(public_network) == 1:
            self.mgmt_network = None
            remove_network = None
            for net_name, net_param in iter(self.test_network_dict.items()):
                if net_name != 'public' and 'router' in net_param \
                        and 'external' in net_param:
                    if not net_param['external']:
                        self.mgmt_network = net_name
                    else:
                        remove_network = net_name
            self.test_network_dict.pop(remove_network)

    def _create_ports_on_networks(self, num_ports=1, **kwargs):
        """Create ports on a test networks for instances

        The method will create a network ports as per test_network dict
        from the external config file.
        The ports creation will loop over the number of specified servers.
        This will allow to call the method once for all instances.

        The ID of the security groups used for the ports creation, removed
        from the kwargs for the later instance creation.
        Note:
        On port creation, tag with port type is set, except to external,
        This will help create_server_with_fip will select required ports
        based on test request filters.

        :param num_ports: The number of loops for ports creation
        :param kwargs
               set_qos: true/false set qos policy during port creation

        :return ports_list: A list of ports lists
        """
        # The "ports_list" holds lists of ports dicts per each instance.
        # Create the number of the nested lists according to the number of the
        # instances.
        ports_list = []
        [ports_list.append([]) for i in range(num_ports)]

        for net_name, net_param in iter(self.test_network_dict.items()):
            if 'skip_srv_attach' in net_param:
                continue
            create_port_body = {'binding:vnic_type': '',
                                'namestart': 'port-smoke',
                                'binding:profile': {}}
            if 'port_type' in net_param:
                create_port_body['binding:vnic_type'] = \
                    net_param['port_type']
                if self.remote_ssh_sec_groups and net_name == \
                        self.mgmt_network:
                    create_port_body['security_groups'] = \
                        [s['id'] for s in self.remote_ssh_sec_groups]
                if 'trusted_vf' in net_param and \
                        net_param['trusted_vf'] and \
                        net_param['port_type'] == 'direct':
                    create_port_body['binding:profile']['trusted'] = True
                if 'switchdev' in net_param and \
                        net_param['switchdev'] and \
                        net_param['port_type'] == 'direct':
                    create_port_body['binding:profile']['capabilities'] = \
                        ['switchdev']

                if len(create_port_body['binding:profile']) == 0:
                    del create_port_body['binding:profile']

                for port_index in range(num_ports):
                    port = self._create_port(network_id=net_param['net-id'],
                                             **create_port_body)
                    # No option to create port with QoS, due to neutron API
                    # Using update port
                    if 'min_qos' in net_param and \
                        net_param['min_qos'] and \
                        net_param['port_type'] == 'direct' and \
                        'set_qos' in kwargs:
                        port_name = data_utils.rand_name('port-min-qos')
                        port_args = {'name': port_name}
                        if kwargs['set_qos']:
                            port_args['qos_policy_id'] = \
                                self.qos_policy_groups['id']
                        self.update_port(port['id'], **port_args)
                    net_var = {'uuid': net_param['net-id'], 'port': port['id']}
                    if 'tag' in net_param:
                        net_var['tag'] = net_param['tag']
                    # Mark port type, as tag
                    else:
                        net_var['tag'] = "{}:{}".format(
                            net_param['port_type'],
                            net_param['provider:physical_network'])
                    # In order to proper map the FIP to the instance,
                    # management network needs to be first in the list of nets.
                    if net_var['tag'] == 'external':
                        ports_list[port_index].insert(0, net_var)
                    else:
                        ports_list[port_index].append(net_var)

        return ports_list

    def _create_port(self, network_id, client=None, namestart='port-quotatest',
                     **kwargs):
        """Port creation for instance

        This Method Overrides Manager::CreatePort to support direct and
        direct ph ports

        :param network_id
        :param client
        :param namestart
        :param kwargs
        """
        kwargs['admin_state_up'] = 'True'
        if not client:
            client = self.ports_client
        name = data_utils.rand_name(namestart)
        result = client.create_port(name=name, network_id=network_id, **kwargs)
        self.assertIsNotNone(result, 'Unable to allocate port')
        port = result['port']
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self.ports_client.delete_port, port['id'])
        return port

    def update_port(self, port_id, **kwargs):
        """update port

        The method, used to update port_body of port.
        kwargs patam should includ additional parameters to be set
        as per the following:
        https://docs.openstack.org/api-ref/network/v2/ \
                ?expanded=update-port-detail#update-port
        :param port_id
        :param kwargs
               qos_policy_id: id of policy to be attached to the port
        """
        ports_client = self.os_admin.ports_client
        ports_client.update_port(port_id, **kwargs)
        self.addCleanup(
            test_utils.call_and_ignore_notfound_exc,
            ports_client.update_port, port_id, qos_policy_id=None)

    def create_server(self, name=None, image_id=None, flavor=None,
                      validatable=False, srv_state=None,
                      wait_on_delete=True, clients=None, **kwargs):
        """This Method Overrides Manager::Createserver to support Gates needs

        :param validatable:
        :param clients:
        :param image_id:
        :param wait_on_delete:
        :param srv_state:
        :param flavor:
        :param name:
        """
        if 'key_name' not in kwargs:
            key_pair = self.create_keypair()
            kwargs['key_name'] = key_pair['name']

        net_id = []
        networks = []
        (CONF.compute_feature_enabled.config_drive
         and kwargs.update({'config_drive': True}))
        if 'networks' in kwargs:
            net_id = kwargs['networks']
            kwargs.pop('networks', None)
        else:
            networks = self.networks_client.list_networks(
                **{'router:external': False})['networks']

        for network in networks:
            net_id.append({'uuid': network['id']})

        server = super(BareMetalManager,
                       self).create_server(name=name,
                                           networks=net_id,
                                           image_id=image_id,
                                           flavor=flavor,
                                           wait_until=srv_state,
                                           **kwargs)
        self.servers.append(server)
        return server

    def create_server_with_fip(self, num_servers=1, use_mgmt_only=False,
                               fip=True, networks=None, srv_state='ACTIVE',
                               raise_on_error=True, **kwargs):
        """Create defined number of the instances with floating ip.

        :param num_servers: The number of servers to boot up.
        :param use_mgmt_only: Boot instances with mgmt net only.
        :param fip: Creation of the floating ip for the server.
        :param networks: List of networks/ports for the servers.
        :param srv_state: The state of the server to expect.
        :param raise_on_error: Raise as error on failed build of the server.
        :param kwargs:
                srv_details: Provide per server override options.
                             Supported options:
                                - flavor (flavor id)
                                - image (image id)
                                - ports_filter (comma separated port types)
                             For example:
                             srv_details = {0: {'flavor': <flavor_id>,
                                                'ports_filter':
                                                    '<port-type-a>,<port-type-b>'}
                                            1: {'flavor': <flavor_id>,
                                                'image': <image_id>}}

        :return: List of created servers
        """
        servers = []
        port = {}

        if not any(isinstance(el, list) for el in networks):
            raise ValueError('Network expect to be as a list of lists')

        override_details = None
        if kwargs.get('srv_details'):
            override_details = kwargs.pop('srv_details')
        for num in range(num_servers):
            kwargs['networks'] = networks[num]

            if override_details:
                if 'flavor' in override_details[num]:
                    kwargs['flavor'] = override_details[num]['flavor']
                if 'image' in override_details[num]:
                    kwargs['image_id'] = override_details[num]['image']
                if 'srv_state' in override_details[num]:
                    kwargs['srv_state'] = override_details[num]['srv_state']
                if 'ports_filter' in override_details[num]:
                    for net in kwargs['networks']:
                        if net['tag'] \
                                not in \
                                override_details[num]['ports_filter']:
                            if net['tag'].split(":")[0] \
                                    not in \
                                    override_details[num]['ports_filter']:
                                kwargs['networks'].remove(net)

            """ If this parameters exist, parse only mgmt network.
            Example live migration can't run with SRIOV ports attached"""
            if use_mgmt_only:
                del (kwargs['networks'][1:])

            LOG.info('Create instance - {}'.format(num + 1))
            servers.append(self.create_server(**kwargs))
        for num, server in enumerate(servers):
            waiters.wait_for_server_status(self.os_admin.servers_client,
                                           server['id'], srv_state,
                                           raise_on_error=raise_on_error)
            LOG.info('The instance - {} is in an {} state'.format(num + 1,
                     srv_state))
            if srv_state == 'ACTIVE':
                port = self.os_admin.ports_client.list_ports(device_id=server[
                    'id'], network_id=networks[num][0]['uuid'])['ports'][0]
            if fip and srv_state == 'ACTIVE':
                server['fip'] = \
                    self.create_floating_ip(server,
                                            port['id'],
                                            self.public_network)[
                        'floating_ip_address']
                LOG.info('The {} fip is allocated to the instance'.format(
                    server['fip']))
            elif srv_state == 'ACTIVE':
                server['fip'] = port['fixed_ips'][0]['ip_address']
                server['network_id'] = networks[num][0]['uuid']
                LOG.info('The {} fixed ip set for the instance'.format(
                    server['fip']))
        return servers

    def create_server_with_resources(self, num_servers=1, num_ports=None,
                                     test=None, **kwargs):
        """The method creates resources and call for the servers method

        The following resources are created:
        - Aggregation
        - Flavor creation / verification
        - Key pair
        - Security groups
        - Test networks
        - Networks ports
        - Cloud init preparation
        - Servers creation
        - Floating ip attachment to the servers
        - QoS attachments to port

        :param num_servers: The number of servers to boot up.
        :param num_ports: The number of ports to the created.
                          Default to (num_servers)
        :param test: Currently executed test. Provide test specific parameters.
        :param kwargs:
                set_qos: true/false create port with qos_policy
                availability_zone: Create and set availability zone
                    zone_name: Name of availability zone (optional)
                    aggr_name: Name of aggregate (optional)
                    hyper_hosts: The list of the hypervisors to be attached
                    aggr_meta: Metadata for aggregate (optional)

                    Example: {'availability_zone': {'hyper_hosts': 'compute0'}}

        :return servers, key_pair
        """
        LOG.info('Creating resources...')

        if num_ports is None:
            num_ports = num_servers

        # Check for the test config file. If not provided, load default.
        if test not in self.test_setup_dict:
            self.create_default_test_config(test)

        # In case resources created externally, set them.
        if self.external_resources_data is not None:
            servers, key_pair = self._organize_external_created_resources(test)
            LOG.info('The resources created by the external tool. '
                     'Continue to the test.')
            return servers, key_pair

        # Create and configure availability zone
        if kwargs.get('availability_zone'):
            avail_zone = kwargs.pop('availability_zone')
            kwargs['availability_zone'] = \
                self.create_and_set_availability_zone(
                    **avail_zone)['availability_zone']

        # Create and configure aggregation zone if specified
        if self.test_setup_dict[test]['aggregate'] is not None:
            aggr_hosts = self.test_setup_dict[test]['aggregate']['hosts']
            aggr_meta = self.test_setup_dict[test]['aggregate']['metadata']
            self.create_and_set_aggregate(test, aggr_hosts, aggr_meta)

        # Flavor creation
        if not kwargs.get('flavor'):
            flavor_check = self.check_flavor_existence(test)
            if flavor_check is False:
                flavor_name = self.test_setup_dict[test]['flavor']
                LOG.info('Flavor {} not found. Creating.'.format(flavor_name))
                try:
                    self.flavor_ref = self.create_flavor(
                        **self.test_flavor_dict[flavor_name])
                except KeyError as exc:
                    err_msg = "Unable to locate {} flavor details for " \
                              "the creation".format(exc)
                    raise Exception(err_msg)

            kwargs['flavor'] = self.flavor_ref

        LOG.info('Creating networks, keypair, security groups, router and '
                 'prepare cloud init.')
        # Key pair creation
        key_pair = self.create_keypair()
        kwargs['key_name'] = key_pair['name']

        # Network, subnet, router and security group creation
        self._create_test_networks()
        self._set_remote_ssh_sec_groups()
        if self.remote_ssh_sec_groups_names:
            kwargs['security_groups'] = self.remote_ssh_sec_groups_names
        ports_list = \
            self._create_ports_on_networks(num_ports=num_ports,
                                           **kwargs)
        # After port creation remove kwargs['set_qos']
        if 'set_qos' in kwargs:
            kwargs.pop('set_qos')
        router_exist = True
        if 'router' in self.test_setup_dict[test]:
            router_exist = self.test_setup_dict[test]['router']
        if router_exist:
            mgmt_subnet_only = False
            if kwargs.get('mgmt_subnet_only'):
                mgmt_subnet_only = True
                kwargs.pop('mgmt_subnet_only')
            self._add_subnet_to_router(mgmt_subnet_only)
        # Prepare cloudinit
        packages = None
        if 'package-names' in self.test_setup_dict[test].keys():
            packages = self.test_setup_dict[test]['package-names']
        kwargs['user_data'] = self._prepare_cloudinit_file(packages)
        servers = []
        if num_servers:
            servers = self.create_server_with_fip(num_servers=num_servers,
                                                  networks=ports_list,
                                                  **kwargs)
        return servers, key_pair

    def _set_remote_ssh_sec_groups(self):
        """Security group creation

        This method create security group except network marked with security
        groups == false in test_networks
        """
        """
        Create security groups [icmp,ssh] for Deployed Guest Image
        """
        mgmt_net = self.mgmt_network
        if not ('sec_groups' in self.test_network_dict[mgmt_net]
                and not self.test_network_dict[mgmt_net]['sec_groups']):
            security_group = self._create_security_group()
            self.remote_ssh_sec_groups_names = \
                [{'name': security_group['name']}]
            self.remote_ssh_sec_groups = [{'name': security_group['name'],
                                           'id': security_group['id']}]

    def _create_security_group(self):
        """Security group creation

        to conform changes in nova clients on microversions>=2.36
        Create security groups and call method create rules
        [icmp,ssh]
        """

        sg_name = data_utils.rand_name(self.__class__.__name__)
        sg_desc = sg_name + " description"
        client = self.security_groups_client
        secgroup = client.create_security_group(
            name=sg_name, description=sg_desc)['security_group']
        self.assertEqual(secgroup['name'], sg_name)
        self.assertEqual(secgroup['description'], sg_desc)
        self.addCleanup(
            test_utils.call_and_ignore_notfound_exc,
            self.security_groups_client.delete_security_group,
            secgroup['id'])

        # Add rules to the security group
        self._create_loginable_secgroup_rule(secgroup['id'])

        return secgroup

    def _create_loginable_secgroup_rule(self, secgroup_id=None):
        """Add secgroups rules

        To conform changes in nova clients on microversions>=2.36
        This method add sg rules with neutron client
        This method find default security group or specific one
        and add icmp and ssh rules
        """
        rule_list = \
            jsonutils.loads(CONF.nfv_plugin_options.login_security_group_rules)
        client = self.security_groups_client
        client_rules = self.security_group_rules_client
        if not secgroup_id:
            sgs = client.list_security_group['security_groups']
            for sg in sgs:
                if sg['name'] == 'default':
                    secgroup_id = sg['id']
                    break

        for rule in rule_list:
            direction = rule.pop('direction')
            client_rules.create_security_group_rule(
                direction=direction,
                security_group_id=secgroup_id,
                **rule)

    def create_floating_ip(self, server, mgmt_port_id, public_network_id):
        """Create floating ip to server

        To conform changes in nova clients on microversions>=2.36
        This method create fip with neutron client
        """
        fip_client = self.floating_ips_client
        floating_ip_args = {
            'floating_network_id': public_network_id,
            'port_id': mgmt_port_id,
            'tenant_id': server['tenant_id']
        }
        floating_ip = \
            fip_client.create_floatingip(**floating_ip_args)['floatingip']
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        fip_client.delete_floatingip,
                        floating_ip['id'])
        return floating_ip

    def check_instance_connectivity(self, ip_addr, user, key_pair):
        """Check connectivity state of the instance

        The function will test the following protocols: ICMP, SSH

        :param ip_addr: The address of the instance
        :param user: Connection user
        :param key_pair: SSH key for the instance connection
        """
        msg = 'Timed out waiting for {} to become reachable'.format(ip_addr)
        self.assertTrue(self.ping_ip_address(ip_addr), msg)
        self.assertTrue(self.get_remote_client(ip_addr, user, key_pair), msg)

    def check_guest_provider_networks(self, servers, key_pair):
        """Check guest provider networks

        This function tests ICMP traffic on all provider networks
        between multiple servers.

        :param servers: List of servers to verify
        :param key-pair: Key pair used to authenticate with server
        """
        # In the current itteration, if only a single server is spawned
        # no pings will be performed.
        # TODO(vkhitrin): In the future, consider pinging default gateway
        if len(servers) == 1:
            LOG.info('Only one server was spawned, no neigbors to ping')
            return True

        for server in servers:
            # Copy servers list to a helper variable
            neighbor_servers = servers[:]
            # Initialize a list of neighbors IPs
            neighbors_ips = []
            # Remove current server from potential server neigbors list
            neighbor_servers.remove(server)
            # Retrieve neighbors IPs from their provier networks
            for neighbor_server in neighbor_servers:
                # Iterate over provider networks for current server and
                # neighbor servers and append potential IP to ping only if
                # both the neighbor and current server are attached to
                # same network
                # Currently it is inefficient to loop this way, consider
                # improving itteration logic
                for neighbor_network in neighbor_server['provider_networks']:
                    for server_network in server['provider_networks']:
                        if neighbor_network['network_id'] == \
                                server_network['network_id']:
                            neighbors_ips.append(
                                neighbor_network['ip_address'])

            ssh_client = self.get_remote_client(server['fip'],
                                                self.instance_user,
                                                key_pair['private_key'])

            hostname = server['name']
            for neighbors_ip in neighbors_ips:
                LOG.info("Guest '{h}' will attempt to "
                         "ping {i}".format(h=hostname, i=neighbors_ip))
                try:
                    ssh_client.icmp_check(neighbors_ip)
                except lib_exc.SSHExecCommandFailed:
                    msg = ("Guest '{h}' failed to ping "
                           "IP '{i}'".format(h=hostname, i=neighbors_ip))
                    raise AssertionError(msg)

                LOG.info("Guest '{h}' successfully was able to ping "
                         "IP '{i}'".format(h=hostname, i=neighbors_ip))

    def get_ovs_port_names(self, servers):
        """This method get ovs port names for each server

        for each server, this method will add mgmt_port and other_port
        values
        :param servers: server list
        return list of ports of each hypervisor
        """
        # get the ports name used for sending/reciving multicast traffic
        # it will be a different port than the management one that will be
        # connected to a switch in which igmp snooping is configured
        port_list = {}
        management_ips = []
        floating_ips = (self.os_admin.floating_ips_client.list_floatingips()
                        ['floatingips'])
        for floating_ip in floating_ips:
            management_ips.append(floating_ip['fixed_ip_address'])
        for server in servers:
            if server['hypervisor_ip'] not in port_list.keys():
                port_list[server['hypervisor_ip']] = []
            ports = self.os_admin.ports_client.list_ports(
                device_id=server['id'])['ports']
            for port in ports:
                ovs_port_name = (port['binding:vif_details']
                                 ['vhostuser_socket'].split('/')[-1])
                if port['fixed_ips'][0]['ip_address'] not in management_ips:
                    server['other_port'] = ovs_port_name
                else:
                    server['mgmt_port'] = ovs_port_name
                port_list[server['hypervisor_ip']].append(ovs_port_name)
        return port_list

    def create_default_test_config(self, test_name=None):
        """This method populate default test configs

        self.test_setup_dict is filled with defaults, if no configuration
        exist for test in network_config

        :param test_name: test_name
        """
        self.assertIsNotNone(test_name,
                             "Please supply mandatory param: test_name")
        if test_name not in self.test_setup_dict:
            self.test_setup_dict[test_name] = \
                {'flavor-id': self.flavor_ref,
                 'router': True, 'aggregate': None}



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from nfv_tempest_plugin.tests.common import shell_utilities as shell_utils
from nfv_tempest_plugin.tests.scenario import base_test
from oslo_log import log as logging
from tempest.common import waiters
from tempest import config

CONF = config.CONF
LOG = logging.getLogger('{} [-] nfv_plugin_test'.format(__name__))


class TestHypervisorScenarios(base_test.BaseTest):
    def __init__(self, *args, **kwargs):
        super(TestHypervisorScenarios, self).__init__(*args, **kwargs)
        self.hypervisor_ip = None

    def setUp(self):
        "Setup a single tenant with an accessible server"
        super(TestHypervisorScenarios, self).setUp()

    def test_hypervisor_reboot(self, test='hypervisor_reboot'):
        """Test functionality of DPDK and SRIOV after hypervisor reboot

        The test will spawn up an instance and then will
        reboot the hypervisor that holds the test instance.
        After hypervisor boot up, the instance will be started
        and tested for the accessability.
        """
        servers, key_pair = self.create_and_verify_resources(test=test)
        LOG.info("Locate instance hypervisor")
        srv_hyper_name = self.os_admin.servers_client.show_server(
            servers[0]['id'])['server']['OS-EXT-SRV-ATTR:host']
        srv_on_hyper = self.hypervisor_client.list_servers_on_hypervisor(
            srv_hyper_name)['hypervisors'][0]['servers']
        LOG.info("Shut down the instances and reboot the hypervisor "
                 "the instance resides on")
        # In order the prevent instances file system corruption,
        # shut down the instance.
        for srv in srv_on_hyper:
            self.servers_client.stop_server(srv['uuid'])
            waiters.wait_for_server_status(self.servers_client, srv['uuid'],
                                           'SHUTOFF')
        shell_utils.run_command_over_ssh(servers[0]['hypervisor_ip'],
                                         "sudo reboot")
        # Reboot of the baremetal hypervisor takes time.
        # In order to not confuse the test, look for the hypervisor status
        # "down" and then "up".
        hyper_rebooted = False
        timeout_start = time.time()
        timeout_end = CONF.nfv_plugin_options.hypervisor_wait_timeout
        while time.time() < timeout_start + timeout_end:
            time.sleep(10)
            hyper_state = self.hypervisor_client.search_hypervisor(
                srv_hyper_name)['hypervisors'][0]['state']
            if 'down' in hyper_state:
                hyper_rebooted = True
                continue
            if hyper_rebooted and 'up' in hyper_state:
                break
        LOG.info("Hypervisor has been rebooted. Booting up the instances.")
        for srv in srv_on_hyper:
            self.servers_client.start_server(srv['uuid'])
            waiters.wait_for_server_status(self.servers_client, srv['uuid'],
                                           'ACTIVE')
        LOG.info("Check instances connectivity")
        for srv in servers:
            self.check_instance_connectivity(ip_addr=srv['fip'],
                                             user=self.instance_user,
                                             key_pair=key_pair['private_key'])
        LOG.info("The hypervisor reboot test passed.")



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# The script sends/receives multicast traffic between multiple hosts.
# For the receivers, the following arguments should be passed:
# receive - Tells the host that it should receive the multicast traffic.
# group - The multicast address, hosts should bind to. Ex. '224.0.0.1'
# port - The port that hosts should wort with. Ex. 10000
# count - The count of packets that host should receive. (optional)
#
# For the sender, the following arguments should be passed:
# send = Tells the host that it should send the multicast traffic.
# group - The multicast address, hosts should bind to. Ex. '224.0.0.1'
# port - The port that hosts should wort with. Ex. 10000
# message - Multicast message to pass to receive hosts.
# count - The count of packets that host should send. (optional)

import argparse
import socket
import struct

parser = argparse.ArgumentParser(
    description='This is multicast send/receive script.')
group = parser.add_mutually_exclusive_group()
group.add_argument('-r', '--receive', help='Receive the multicast traffic',
                   action='store_true', required=False)
group.add_argument('-s', '--send', help='Send the multicast traffic',
                   action='store_true', required=False)
parser.add_argument('-g', '--group', help='Multicast bind group',
                    required=True)
parser.add_argument('-p', '--port', help='Multicast port', required=True)
parser.add_argument('-m', '--message', help='Multicast message',
                    required=False)
parser.add_argument('-c', '--count', help='Count of packets to send/receive',
                    required=False)
args = parser.parse_args()

sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
if args.receive:
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((args.group, int(args.port)))
    mreq = struct.pack("4sl", socket.inet_aton(args.group), socket.INADDR_ANY)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

    count = 1
    while True:
        print(sock.recv(1024))
        if args.count:
            if int(args.count) == count:
                break
            count += 1

if args.send:
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 32)

    count = 1
    while True:
        sock.sendto(args.message, (args.group, int(args.port)))
        if args.count:
            if int(args.count) == count:
                break
            count += 1



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# The script attempts to generate ICMP and MPLS traffic as well as sniffing
# the traffic on defined interfaces.

import argparse
import logging
import sys
import time

from scapy.contrib.mpls import MPLS
from scapy.layers.inet import ICMP
from scapy.layers.inet import IP
from scapy.layers.l2 import Dot1Q
from scapy.layers.l2 import Ether
from scapy.packet import Raw
from scapy.sendrecv import sendp
from scapy.sendrecv import sniff


parser = argparse.ArgumentParser(
    description='This is scapy send/receive traffic script.')
group = parser.add_mutually_exclusive_group()
group.add_argument('--icmp', help='Send/receive ICMP traffic', required=False,
                   action='store_true')
group.add_argument('--mpls', help='Send/receive MPLS traffic', required=False,
                   action='store_true')
group.add_argument('--sniff', help='Sniff incomming traffic',
                   action='store_true', required=False)
parser.add_argument('-i', '--interface', help='Use specified interface',
                    required=True)
parser.add_argument('-c', '--count', help='Number of packets to send/receive',
                    default=5, type=int)
parser.add_argument('-t', '--timeout', help='Timeout for the sniffing wait',
                    default=120, type=int)
parser.add_argument('--src-mac', help='Source mac address', required=False)
parser.add_argument('--dst-mac', help='Destination mac address',
                    required=False)
parser.add_argument('--src-ip', help='Source ip address', required=False)
parser.add_argument('--dst-ip', help='Destination ip address', required=False)
parser.add_argument('--iface-vlan', help='The vlan that will be set on the '
                                         'virtual interface',
                    type=int, required=False)
parser.add_argument('--test-vlan', help='The vlan that will be used for '
                                        'packets send',
                    type=int, required=False)
parser.add_argument('--raw-msg', help='Include raw massage into the packet',
                    required=False)
parser.add_argument('--keep-sniff', help='Keep sniffing for the packets and'
                                         'do not exit',
                    required=False, action='store_true')
args = parser.parse_args()

logging.basicConfig(filename='/tmp/scapy_traffic.log', filemode='a',
                    format='%(asctime)s %(name)s %(levelname)s: %(message)s',
                    datefmt="%h %d %H:%M:%S", level=logging.INFO)
handler = logging.StreamHandler(sys.stdout)
logger = logging.getLogger("ScapyTraffic")
logger.addHandler(handler)


class ScapyTraffic(object):

    def __init__(self):
        self.icmp = args.icmp
        self.mpls = args.mpls
        self.sniff = args.sniff
        self.iface = args.interface
        self.count = args.count
        self.timeout = args.timeout
        self.src_mac = args.src_mac
        self.dst_mac = args.dst_mac
        self.src_ip = args.src_ip
        self.dst_ip = args.dst_ip
        self.iface_vlan = args.iface_vlan
        self.test_vlan = args.test_vlan
        self.raw_msg = args.raw_msg

    def sniff_catch(self):
        return sniff(iface=self.iface, count=self.count, timeout=self.timeout)

    @staticmethod
    def sniff_parse(sniff_output=None):
        if sniff_output is not None:
            pkt_count = len(sniff_output)
            if 'MPLS' in sniff_output[0] and sniff_output[0].type == 34887:
                pkt_type = 'mpls'
            elif 'ICMP' in sniff_output[0] and \
                    sniff_output[0].getlayer(ICMP).type == 8:
                pkt_type = 'icmp'
            else:
                pkt_type = 'unidentified'
            custom_string = sniff_output[0].load if sniff_output[0].load \
                else False
            return {'pkt_count': pkt_count, 'pkt_type': pkt_type,
                    'custom_string': custom_string}
        return False

    def send_icmp(self):
        icmp_eth = Ether(src=self.src_mac, dst=self.dst_mac)
        icmp_id = Dot1Q(vlan=self.iface_vlan, id=3, prio=2) / Dot1Q(
            vlan=self.test_vlan, id=3, prio=2)
        icmp_ip = IP(src=self.src_ip, dst=self.dst_ip)
        icmp_icmp = ICMP()
        icmp_frame = icmp_eth / icmp_id / icmp_ip / icmp_icmp

        send_icmp_log = icmp_frame.sprintf("Packets send from mac: "
                                           "%Ether.src%, ip: %IP.src% to mac: "
                                           "%Ether.dst%, ip: %IP.dst% using "
                                           "vlans %Dot1Q.vlan% and")
        try:
            logger.info("Send ICMP traffic from the {iface} interface with "
                        "the following details: \n{packet_details} "
                        "{second_vlan}".format(iface=self.iface,
                                               packet_details=send_icmp_log,
                                               second_vlan=self.test_vlan))
            sendp(icmp_frame, iface=self.iface, count=self.count)
        except Exception as e:
            logger.info("The ICMP traffic send failed due"
                        " to: {}".format(e.message))
        return

    def send_mpls(self):
        mpls_eth = Ether(src=self.src_mac, dst=self.dst_mac, type=0x8847)
        mpls_lables = MPLS(label=16, s=0, ttl=255) / MPLS(
            label=18, s=0, ttl=255) / MPLS(
            label=18, s=0, ttl=255) / MPLS(
            label=16, s=1, ttl=255)
        mpls_ip = IP(src=self.src_ip, dst=self.dst_ip)
        mpls_icmp = ICMP(type="echo-request")
        mpls_raw = Raw(load=self.raw_msg)
        mpls_frame = mpls_eth / mpls_lables / mpls_ip / mpls_icmp / mpls_raw

        send_mpls_log = mpls_frame.sprintf("Packets send from mac: "
                                           "%Ether.src%, ip: %IP.src% to mac: "
                                           "%Ether.dst%, ip: %IP.dst% with the"
                                           " following raw message: "
                                           "%Raw.load%")
        try:
            logger.info("Send MPLS traffic from the {iface} interface with "
                        "the following details: \n{packet_details}"
                        .format(iface=self.iface,
                                packet_details=send_mpls_log))
            sendp(mpls_frame, iface=self.iface, count=self.count)
        except Exception as e:
            logger.info("The MPLS traffic send failed due"
                        " to: {}".format(e.message))
        return


def main():
    scap = ScapyTraffic()
    logger.info("Start scapy traffic script")

    if scap.sniff:
        timeout_start = time.time()
        while time.time() < timeout_start + args.timeout:
            logger.info("Start sniff on the {} "
                        "interface".format(args.interface))
            sniff_output = scap.sniff_catch()
            sniff_parse_out = None
            if sniff_output:
                sniff_parse_out = scap.sniff_parse(sniff_output)
                logger.info("Sniff parsed output: {}".format(sniff_parse_out))
            else:
                logger.error("The sniffer was unable to sniff any packet")
            if not args.keep_sniff and sniff_parse_out:
                return sniff_parse_out

    if scap.icmp:
        return scap.send_icmp()

    if scap.mpls:
        return scap.send_mpls()


if __name__ == '__main__':
    main()



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nfv_tempest_plugin.tests.scenario import base_test
from oslo_log import log as logging
from tempest.common import waiters
from tempest import config

CONF = config.CONF
LOG = logging.getLogger('{} [-] nfv_plugin_test'.format(__name__))


class TestAdvancedScenarios(base_test.BaseTest):
    def __init__(self, *args, **kwargs):
        super(TestAdvancedScenarios, self).__init__(*args, **kwargs)

    def setUp(self):
        """Set up a single tenant with an accessible server

        If multi-host is enabled, save created server uuids.
        """
        super(TestAdvancedScenarios, self).setUp()

    def test_numa_aware_vswitch(self, test='numa_aware_vswitch'):
        """Test proper vcpu allocation according to numa aware vswitch config

        The numa aware vswitch allows to configure the allocation of the
        instance vcpu by mapping the physnet to the proper numa node.

        Note!
        - The test require NUMATopologyFilter to be set.

        Note! - The test suit only for OSP Rocky version and above, since the
                numa aware vswitch feature was implemented only in OSP Stein
                version and backported to OSP Rocky.
        """
        LOG.info('Starting numa aware vswitch test.')
        LOG.info('Create resources for the test.')
        hyper = self.os_admin.hypervisor_client.list_hypervisors()[
            'hypervisors']
        kwargs = {'availability_zone': {'zone_name': 'numa_aware_avail_zone',
                                        'hyper_hosts':
                                            [hyper[0]['hypervisor_hostname']]}}
        _, key_pair = self.create_and_verify_resources(test=test,
                                                       num_servers=0,
                                                       **kwargs)
        LOG.info('Gather numa aware and non numa aware network details.')
        numas_phys = self.locate_ovs_physnets()
        if not numas_phys.get('numa_aware_net'):
            raise ValueError('Numa aware physnet configuration is missing')
        numa_net = self.locate_numa_aware_networks(numas_phys)

        resources = self.list_available_resources_on_hypervisor(
            hyper[0]['hypervisor_hostname'])
        flavor_vcpu = self.os_primary.flavors_client.show_flavor(
            self.flavor_ref)['flavor']['vcpus']
        srv_num_to_boot = resources['pcpu_free_per_numa'] // flavor_vcpu
        net_id = []
        for _ in range(srv_num_to_boot):
            net_id.append([{'uuid': numa_net}])
        kwargs = {'security_groups': self.remote_ssh_sec_groups_names,
                  'availability_zone': 'numa_aware_avail_zone',
                  'key_name': key_pair['name']}
        LOG.info('Booting instances on numa node 0. Expect to succeed.')
        numa_aware_srv = \
            self.create_server_with_fip(num_servers=srv_num_to_boot,
                                        flavor=self.flavor_ref,
                                        networks=net_id, **kwargs)
        for srv in numa_aware_srv:
            LOG.info('Instance details: fip: {}, instance_id: {}'.format(
                srv['fip'], srv['id']))
            srv['hypervisor_ip'] = self._get_hypervisor_ip_from_undercloud(
                **{'shell': CONF.nfv_plugin_options.undercloud_rc_file,
                   'server_id': srv['id']})[0]
            self.assertNotEmpty(srv['hypervisor_ip'],
                                "_get_hypervisor_ip_from_undercloud "
                                "returned empty ip list")

        LOG.info('Booting up another numa aware instance. Expect to fail.')
        fail_srv = self.create_server_with_fip(flavor=self.flavor_ref,
                                               srv_state='ERROR',
                                               raise_on_error=False,
                                               networks=net_id,
                                               **kwargs)
        fail_srv_state = self.os_primary.servers_client.show_server(
            fail_srv[0]['id'])['server']
        self.assertEqual(fail_srv_state['status'], 'ERROR')
        LOG.info('Check vcpu placement of numa aware instances.')
        [self.match_vcpu_to_numa_node(
            srv, srv['hypervisor_ip'],
            numa_node=numas_phys['numa_aware_net']['numa_node'])
         for srv in numa_aware_srv]
        srv_list = [srv['id'] for srv in numa_aware_srv]

        non_numa_skip_log = ('Skip "non numa aware" test phase as '
                             '"non numa aware" network was not found')
        if numas_phys.get('non_numa_aware_net'):
            non_numa_net = None
            # Exclude external network to avoid issue with booting instance
            # and assign external network and internal to the instance.
            non_numa_net = self.networks_client.list_networks(
                **{'provider:physical_network': numas_phys[
                    'non_numa_aware_net'],
                   'router:external': False})['networks']
            if non_numa_net:
                LOG.info('Test non numa aware network')
                non_numa_net = non_numa_net[0]['id']
                net_id = [[{'uuid': non_numa_net}]]
                LOG.info('Booting an non numa aware instance.'
                         'Expect to success.')
                non_numa_srv = \
                    self.create_server_with_fip(flavor=self.flavor_ref,
                                                networks=net_id, **kwargs)
                srv_list.append(non_numa_srv[0]['id'])
            else:
                LOG.warn(non_numa_skip_log)
        else:
            LOG.warn(non_numa_skip_log)

        LOG.info('Ensure all the instances are on the same hypervisor node.')
        host = [self.os_admin.servers_client.show_server(hp)['server']
                ['OS-EXT-SRV-ATTR:hypervisor_hostname'] for hp in srv_list]
        host = list(set(host))
        if len(host) > 1:
            raise ValueError("The instances should reside on a single "
                             "hypervisor. Use availability zone to reach "
                             "that state.")
        LOG.info('Migrate the instance')
        self.os_admin.servers_client.live_migrate_server(
            server_id=numa_aware_srv[0]['id'], block_migration=True,
            host=hyper[1]['hypervisor_hostname'], force=True)
        waiters.wait_for_server_status(self.servers_client,
                                       numa_aware_srv[0]['id'], 'ACTIVE')
        LOG.info('Verify instance connectivity after the cold migration.')
        self.check_instance_connectivity(ip_addr=numa_aware_srv[0]['fip'],
                                         user=self.instance_user,
                                         key_pair=key_pair['private_key'])
        second_hyper = self._get_hypervisor_ip_from_undercloud(
            **{'shell': CONF.nfv_plugin_options.undercloud_rc_file,
               'server_id': numa_aware_srv[0]['id']})[0]
        self.assertNotEqual(numa_aware_srv[0]['hypervisor_ip'], second_hyper,
                            'The instance was not able to migrate to '
                            'another hypervisor')
        LOG.info('The {} instance has been migrated to the {} hypervisor'
                 .format(numa_aware_srv[0]['id'], second_hyper))
        LOG.info('Cold migration passed.')
        LOG.info('Numa aware test completed.')

    def test_pinned_srv_live_migration(self, test='pinned_srv_live_migration'):
        """Test live migration of pinned instances.

        The test performs the following actions:
        - Boot the cpu pinned instance on the first hypervisor
        - Live migrate the cpu pinned instance to the second hypervisor
          Expect live migration to success
        - Boot seconds pinned instance on the first hypervisor
        - Live migrate the first instance back to the first hypervisor
        - Verify by the virsh xml that the first vm was rescheduled on the cpu.
        """
        LOG.info('Pinned instance live migration test')
        srv1, key_pair = self.create_and_verify_resources(test=test,
                                                          use_mgmt_only=True)
        srv1_vcpus_before_migration = \
            self.get_instance_vcpu(srv1[0], srv1[0]['hypervisor_ip'])
        LOG.info('The cores of {srv} instance on the {hyper} hypervisor are '
                 '{cores}'.format(srv=srv1[0]['id'],
                                  hyper=srv1[0]['hypervisor_ip'],
                                  cores=srv1_vcpus_before_migration))

        LOG.info('Live migrate the instance to the second hypervisor')
        self.os_admin.servers_client.live_migrate_server(
            server_id=srv1[0]['id'], block_migration=True, host=None)
        waiters.wait_for_server_status(self.servers_client, srv1[0]['id'],
                                       'ACTIVE')
        self.check_instance_connectivity(ip_addr=srv1[0]['fip'],
                                         user=self.instance_user,
                                         key_pair=key_pair['private_key'])
        LOG.info('Verify the migration succeeded')
        second_hyper = self._get_hypervisor_ip_from_undercloud(
            **{'shell': CONF.nfv_plugin_options.undercloud_rc_file,
               'server_id': srv1[0]['id']})[0]
        self.assertNotEqual(srv1[0]['hypervisor_ip'], second_hyper,
                            'The instance was not able to migrate to '
                            'another hypervisor')
        LOG.info('The {} instance has been migrated to the {} hypervisor'
                 .format(srv1[0]['id'], second_hyper))

        mgmt_net = self.mgmt_network
        mgmt_net_id = [[{'uuid': self.test_network_dict[mgmt_net]['net-id']}]]
        kwargs = {'security_groups': self.remote_ssh_sec_groups_names,
                  'key_name': key_pair['name']}
        srv2 = self.create_server_with_fip(flavor=self.flavor_ref,
                                           networks=mgmt_net_id, **kwargs)
        self.check_instance_connectivity(ip_addr=srv2[0]['fip'],
                                         user=self.instance_user,
                                         key_pair=key_pair['private_key'])
        srv2[0]['hypervisor_ip'] = self._get_hypervisor_ip_from_undercloud(
            **{'shell': CONF.nfv_plugin_options.undercloud_rc_file,
               'server_id': srv2[0]['id']})[0]
        LOG.info('Boot second instance {} on the {} hypervisor'
                 .format(srv2[0]['id'], srv2[0]['hypervisor_ip']))
        srv2_vcpus = self.get_instance_vcpu(srv2[0], srv1[0]['hypervisor_ip'])
        LOG.info('The cores of {} instance on the {} hypervisor are {}'.format(
            srv2[0]['id'], srv2[0]['hypervisor_ip'], srv2_vcpus))

        LOG.info('Live migrate srv1 back to the first hypervisor')
        self.os_admin.servers_client.live_migrate_server(
            server_id=srv1[0]['id'], block_migration=True, host=None)
        waiters.wait_for_server_status(self.servers_client, srv1[0]['id'],
                                       'ACTIVE')
        self.check_instance_connectivity(ip_addr=srv1[0]['fip'],
                                         user=self.instance_user,
                                         key_pair=key_pair['private_key'])
        first_hyper = self._get_hypervisor_ip_from_undercloud(
            **{'shell': CONF.nfv_plugin_options.undercloud_rc_file,
               'server_id': srv1[0]['id']})[0]
        self.assertEqual(srv1[0]['hypervisor_ip'], first_hyper,
                         'The {} instance was not migrated back to the {} '
                         'hypervisor'. format(srv1[0]['id'], first_hyper))
        srv1_vcpus_after_migration = \
            self.get_instance_vcpu(srv1[0], srv1[0]['hypervisor_ip'])
        LOG.info('The cores of {} instance on the {} hypervisor after '
                 'migration are {}'.format(srv1[0]['id'], first_hyper,
                                           srv1_vcpus_after_migration))

        LOG.info('Ensure srv2 uses released cores of migrated srv1 instance')
        self.assertEqual(srv1_vcpus_before_migration, srv2_vcpus,
                         'The cores are not equal: {srv1} - {srv2}. '
                         'The second boot instance should take the same cores'
                         ' released by the migrated instance'.format(
                             srv1=srv1_vcpus_before_migration,
                             srv2=srv2_vcpus))
        LOG.info('Ensure that srv1 migrated back to the first hypervisor, '
                 'rescheduled its cores')
        self.assertNotEqual(srv2_vcpus, srv1_vcpus_after_migration,
                            'The cores are equal: {srv1_cpu} - {srv2_cpu}. '
                            'No core re-schedule detected!!! Once {srv1} '
                            'instance migrated back to {hyper} hypervisor, '
                            'its cores should differ from the cores before '
                            'the migration'
                            .format(srv1_cpu=srv1_vcpus_after_migration,
                                    srv2_cpu=srv2_vcpus,
                                    srv1=srv1[0]['id'],
                                    hyper=srv1[0]['hypervisor_ip']))
        LOG.info('The pinned instance live migration test passed')

    def test_pinned_and_non_pinned_srv(self, test='pinned_and_non_pinned_srv'):
        """Test pinned and non pinned instances on the same compute

        The test performs the following actions:
        - Boot pinned instance (using specific flavor)
        - Boot non pinned instance (using specific flavor) on the same host
        - Ensure the instances booted on the same hypervisor
        - Takes the allocated cpu for the intstances
        - Takes the dedicated and shared cpu set from hypervisor
        - Compares between them to ensure that instances uses proper cpu
        """
        LOG.info('Pinned and non pinned instances test')
        LOG.info('Create flavors for pinned and non pinned instances')
        flavors_id = []
        flavors = [{'name': 'flavor_dedicated', 'vcpus': 4,
                    'extra_specs': {'hw:mem_page_size': "large",
                                    'resources:PCPU': '4',
                                    'hw:emulator_threads_policy': 'share'}},
                   {'name': 'flavor_shared', 'vcpus': 2,
                    'extra_specs': {'hw:mem_page_size': "large",
                                    'resources:VCPU': '2',
                                    'hw:emulator_threads_policy': 'share'}}]
        for flavor in flavors:
            flavors_id.append(self.create_flavor(**flavor))
        srv_details = {'srv_details': {0: {'flavor': flavors_id[0]},
                                       1: {'flavor': flavors_id[1]}}}
        hyper = self.hypervisor_client.list_hypervisors()['hypervisors']
        kwargs = {'availability_zone':
                  {'hyper_hosts': [hyper[0]['hypervisor_hostname']]}}
        kwargs.update(srv_details)
        servers, _ = self.create_and_verify_resources(test=test, num_servers=2,
                                                      use_mgmt_only=True,
                                                      **kwargs)
        LOG.info('Ensure all the instances are on the same hypervisor node.')
        srv_id = []
        for srv in servers:
            srv_id.append(srv['id'])
        hyper = [self.os_admin.servers_client.show_server(hp)['server']
                 ['OS-EXT-SRV-ATTR:hypervisor_hostname'] for hp in srv_id]
        hyper = list(set(hyper))
        if len(hyper) > 1:
            raise ValueError("The instances should reside on a single "
                             "hypervisor. Use aggregate to reach that state.")
        LOG.info('Check instances vcpu placement')
        srv_dedicated = self.get_instance_vcpu(servers[0],
                                               servers[0]['hypervisor_ip'])
        srv_shared = self.get_instance_vcpu(servers[1],
                                            servers[1]['hypervisor_ip'])
        LOG.info('Take the dedicated and share cpu set from hypervisor')
        dedicated_set, shared_set = self.locate_dedicated_and_shared_cpu_set()
        LOG.info('CHeck results')
        test_results = []
        if not all(cpu in shared_set for cpu in srv_shared):
            test_results.append('The vcpu\'s used by the shared instance - {} '
                                'does not exist in shared '
                                'set - {}.'.format(srv_shared, shared_set))
        if not all(cpu in dedicated_set for cpu in srv_dedicated):
            test_results.append('The vcpu\'s used by the dedicated instance '
                                '- {} does not exist in dedicated set '
                                '- {}.'.format(srv_dedicated, dedicated_set))
        self.assertEmpty(test_results, test_results)
        LOG.info('Pinned and non pinned instances test passed')



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import setuptools

# In python < 2.7.4, a lazy loading of package `pbr` will break
# setuptools if some other modules registered functions in `atexit`.
# solution from: http://bugs.python.org/issue15881#msg170215
try:
    import multiprocessing  # noqa
except ImportError:
    pass

setuptools.setup(
    setup_requires=['pbr>=1.9', 'setuptools>=17.1'],
    pbr=True
)



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import os.path
import paramiko
import re
import textwrap
import time
import xml.etree.ElementTree as ELEMENTTree
import yaml

from nfv_tempest_plugin.tests.common import shell_utilities as shell_utils
from oslo_log import log
from oslo_serialization import jsonutils
from tempest import config
from tempest.lib import exceptions as lib_exc
"""Python 2 and 3 support"""
from six.moves import StringIO
from six.moves.urllib.parse import urlparse

CONF = config.CONF
LOG = log.getLogger('{} [-] nfv_plugin_test'.format(__name__))


class ManagerMixin(object):
    def read_external_config_file(self):
        """This Method reads network_config.yml

        Reads config data and assign it to dictionaries
        """
        with open(CONF.nfv_plugin_options.external_config_file, 'r') as f:
            self.external_config = yaml.safe_load(f)

        if not os.path.exists(
                CONF.nfv_plugin_options.external_resources_output_file):
            """Hold flavor, net and images lists"""
            # TODO(read and parse to util move to util)
            networks = self.networks_client.list_networks()['networks']
            flavors = self.flavors_client.list_flavors()['flavors']
            images = self.image_client.list_images()['images']

            if 'networks' in self.external_config:
                """
                Iterate over networks mandatory vars in external_config are:
                port_type, gateway_ip
                """
                for net in self.external_config['networks']:
                    self.test_network_dict[net['name']] = {'port_type': net[
                        'port_type'], 'gateway_ip': net['gateway_ip']}
                    """
                    Check for existence of optionals vars:
                    router_name, external.
                    """
                    if 'external' in net:
                        self.test_network_dict[net['name']]['external'] = net[
                            'external']
                    if 'router_name' in net:
                        self.test_network_dict[net['name']]['router'] = net[
                            'router_name']

                # iterate networks
                for net in iter(self.test_network_dict.keys()):
                    for network in networks:
                        if network['name'] == net:
                            self.test_network_dict[net]['net-id'] = \
                                network['id']

        # Insert here every new parameter.
        for test in self.external_config['tests-setup']:
            self.test_setup_dict[test['name']] = {}
            self.test_setup_dict[test['name']]['config_dict'] = {}
            if 'flavor' in test and test['flavor'] is not None:
                self.test_setup_dict[test['name']]['flavor'] = test['flavor']
            if 'package-names' in test and test['package-names'] is not None:
                self.test_setup_dict[test['name']]['package-names'] = \
                    test['package-names']
            if 'image' in test and test['image'] is not None:
                self.test_setup_dict[test['name']]['image'] = \
                    test['image']
            if 'router' in test and test['router'] is not None:
                self.test_setup_dict[test['name']]['router'] = \
                    test['router']
            if 'service-names' in test and test['service-names'] is not None:
                self.test_setup_dict[test['name']]['service-names'] = \
                    test['service-names']
            if 'tuned-profile' in test and test['tuned-profile'] is not None:
                self.test_setup_dict[test['name']]['tuned-profile'] = \
                    test['tuned-profile']
            if 'mtu' in test and test['mtu'] is not None:
                self.test_setup_dict[test['name']]['mtu'] = \
                    test['mtu']
            if 'emulatorpin_config' in test and test['emulatorpin_config'] \
                    is not None:
                for item in test['emulatorpin_config']:
                    for key, value in iter(item.items()):
                        if not value:
                            raise ValueError('The {0} configuration is '
                                             'required for the emulatorpin '
                                             'test, but currently empty.'
                                             .format(key))
                epin_str = jsonutils.dumps(test['emulatorpin_config'])
                self.test_setup_dict[test['name']]['config_dict'] = \
                    jsonutils.loads(epin_str)
            if 'rx_tx_config' in test and test['rx_tx_config'] is not None:
                for item in test['rx_tx_config']:
                    for key, value in iter(item.items()):
                        if not value:
                            raise ValueError('The {0} configuration is '
                                             'required for the tx/tx test, '
                                             'but currently empty.'
                                             .format(key))
                rx_tx_str = jsonutils.dumps(test['rx_tx_config'])
                self.test_setup_dict[test['name']]['config_dict'] = \
                    jsonutils.loads(rx_tx_str)
            if 'bonding_config' in test and test['bonding_config'] is not None:
                for item in test['bonding_config']:
                    for key, value in iter(item.items()):
                        if not value:
                            raise ValueError('The {0} configuration is '
                                             'required for the bondig '
                                             'test, but currently empty.'
                                             .format(key))
                bonding_str = jsonutils.dumps(test['bonding_config'])
                test_setup_dict = self.test_setup_dict[test['name']]
                test_setup_dict['config_dict']['bonding_config'] = \
                    jsonutils.loads(bonding_str)
            if 'igmp_config' in test and test['igmp_config'] is not None:
                for item in test['igmp_config']:
                    for key, value in iter(item.items()):
                        if not value:
                            raise ValueError('The {0} configuration is '
                                             'required for the igmp '
                                             'test, but currently empty.'
                                             .format(key))
                igmp_str = jsonutils.dumps(test['igmp_config'])
                test_setup_dict = self.test_setup_dict[test['name']]
                test_setup_dict['config_dict']['igmp_config'] = \
                    jsonutils.loads(igmp_str)

            self.test_setup_dict[test['name']]['aggregate'] = \
                test.get('aggregate')
            self.test_setup_dict[test['name']]['vlan_config'] = \
                test.get('vlan_config')

            if 'offload_nics' in test and test['offload_nics'] is not None:
                self.test_setup_dict[test['name']]['offload_nics'] = \
                    test['offload_nics']

            if 'qos_rules' in test and test['qos_rules'] is not None:
                self.test_setup_dict[test['name']]['qos_rules'] = \
                    jsonutils.loads(jsonutils.dumps(test['qos_rules']))

        if not os.path.exists(
                CONF.nfv_plugin_options.external_resources_output_file):
            # iterate flavors_id
            for test, test_param in iter(self.test_setup_dict.items()):
                if 'flavor' in test_param:
                    for flavor in flavors:
                        if test_param['flavor'] == flavor['name']:
                            self.test_setup_dict[test]['flavor-id'] = flavor[
                                'id']

            # iterate image_id
            for test, test_param in iter(self.test_setup_dict.items()):
                if 'image' in test_param:
                    for image in images:
                        if test_param['image'] == image['name']:
                            self.test_setup_dict[test]['image-id'] = \
                                image['id']

        # iterate flavors parameters
        if 'test-flavors' in self.external_config:
            for flavor in self.external_config['test-flavors']:
                self.test_flavor_dict[flavor['name']] = flavor

        if 'test_instance_repo' in self.external_config:
            self.test_instance_repo = self.external_config[
                'test_instance_repo']

        if 'user_data' in CONF.nfv_plugin_options:
            self.user_data = CONF.nfv_plugin_options.user_data
            self.assertTrue(os.path.exists(self.user_data),
                            "Specified user_data file can't be read")
        # Update the floating IP configuration (enable/disable)
        self.fip = self.external_config.get('floating_ip', True)

    def get_osp_release(self, hypervisor=None):
        """Gather OSP release

        Takes the OSP release from the hypervisor
        :param hypervisor: Ip of the hypervisor to work on (optional)
        :return OSP version integer
        """
        if not hypervisor:
            hyper_kwargs = {'shell':
                            CONF.nfv_plugin_options.undercloud_rc_file}
            hypervisor = self._get_hypervisor_ip_from_undercloud(
                **hyper_kwargs)[0]
        ver = shell_utils.\
            run_command_over_ssh(hypervisor, 'cat /etc/rhosp-release')
        if ver == '':
            ver = shell_utils.run_command_over_ssh(
                hypervisor,
                'cat /var/lib/rhos-release/latest-installed')
        return int(re.findall(r'\d+', ver)[0])

    def check_flavor_existence(self, testname):
        """Check test specific flavor existence.

        :param testname: value - The name of the running test.
        """
        if 'flavor' and 'flavor-id' in self.test_setup_dict[testname]:
            self.flavor_ref = self.test_setup_dict[testname]['flavor-id']
            return True
        return False

    def _get_dumpxml_instance_data(self, server, hypervisor):
        """Get dumpxml data from the instance

        :param server: Server name
        :param hypervisor: Hypervisor that hold the instance

        :return dumpxml_string
        """

        server_details = \
            self.os_admin.servers_client.show_server(server['id'])['server']
        osp_release = self.get_osp_release(hypervisor)
        # If OSP version is 16, use podman container to retrieve instance XML
        if osp_release >= 16:
            cmd = ('sudo podman exec -it nova_libvirt virsh -c '
                   'qemu:///system dumpxml {}')
        else:
            cmd = 'sudo virsh -c qemu:///system dumpxml {}'
        get_dumpxml = \
            cmd.format(server_details['OS-EXT-SRV-ATTR:instance_name'])
        dumpxml_data = shell_utils.\
            run_command_over_ssh(hypervisor, get_dumpxml)
        dumpxml_string = ELEMENTTree.fromstring(dumpxml_data)

        return dumpxml_string

    def get_instance_vcpu(self, instance, hypervisor):
        """Get a list of vcpu cores used by the instance

        :param instance
        :param hypervisor
        :return list of instance vcpu
        """
        dumpxml_string = self._get_dumpxml_instance_data(instance, hypervisor)
        vcpupin = dumpxml_string.findall('./cputune/vcpupin')
        vcpu_list = [(vcpu.get('cpuset'))
                     for vcpu in vcpupin if vcpu is not None]
        vcpu_total_list = list()
        for vcpu in vcpu_list:
            if ',' in vcpu or '-' in vcpu:
                sep = ',' if ',' in vcpu else '-'
                split_list = vcpu.split(sep) if sep in vcpu else None
                if '-' in vcpu:
                    split_list = list(range(int(split_list[0]),
                                            int(split_list[1]) + 1))
                vcpu_total_list.extend(split_list)
        if vcpu_total_list:
            vcpu_list = vcpu_total_list
        vcpu_final_list = [int(vcpu) for vcpu in vcpu_list]
        return list(set(vcpu_final_list))

    def match_vcpu_to_numa_node(self, instance, hypervisor, numa_node='0'):
        """Verify that provided vcpu list resides within the specified numa

        :param instance: The instance that should check the vcpu on
        :param hypervisor: Check cores on specified hypervisor (ip address)
        :param numa_node: Specify the numa node to check vcpu in
                          String choices: 0, 1, mix
        """
        # In case of mix topology checking only node0 and verifying
        # dumpxml_vcpu_list > hyper_vcpu_list
        mix_mode = True if numa_node == 'mix' else False

        dumpxml_vcpu_list = self.get_instance_vcpu(instance, hypervisor)
        bash_array = " ".join(['{}'.format(x) for x in dumpxml_vcpu_list])
        cmd = '''
        array=( {cpu_list} ); for i in "${{array[@]}}";do
        if [ -d /sys/devices/system/cpu/cpu$i/node{cell} ];then
        echo $i; fi; done'''.format(cell=numa_node, cpu_list=bash_array)
        hyper_vcpu_list = shell_utils.\
            run_command_over_ssh(hypervisor, cmd).split()
        hyper_vcpu_list = [int(core) for core in hyper_vcpu_list]
        # !!! In case of Mix search for hyper_vcpu_list smaller than vcpu_list
        if mix_mode:
            self.assertIsNot(len(hyper_vcpu_list), len(dumpxml_vcpu_list),
                             'Number of mix vCPUs on numa node {numa} is equal'
                             ' to config {result}'.format(
                                 numa=numa_node, result=hyper_vcpu_list))
        else:
            self.assertEqual(hyper_vcpu_list, dumpxml_vcpu_list,
                             'Number of vCPUs on numa node {numa} does not '
                             'match to config {result}'.format(
                                 numa=numa_node, result=hyper_vcpu_list))

    def _check_numa_from_dumpxml(self, server, hypervisor):
        """Instance number of cells check

        This method checks the number of cells within the provided dumpxml data

        :param server
        :param hypervisor
        """
        dumpxml_string = self._get_dumpxml_instance_data(server, hypervisor)

        dumpxml = dumpxml_string.findall('cpu')[0]
        for i in dumpxml.findall('topology')[0].items():
            if i[0] == 'sockets':
                # change to 2
                self.assertEqual(i[1], '1')
                print(i[0])
        count = 0
        for i in dumpxml.findall('numa')[0].findall('cell'):
            # change memory to 1572864
            if (('id', '0') in i.items() and (
                    ('memory', '2097152')) in i.items()):
                count += 1
            # change cell id to 1 memory to 524288
            if (('id', '1') in i.items() and (
                    ('memory', '2097152')) in i.items()):
                count += 1
        self.assertEqual(count, '2')

    def _check_emulatorpin_from_dumpxml(self, server, hypervisor):
        """Emulatorpin configuration on the instance

        :param server
        :param hypervisor
        """

        dumpxml_string = self._get_dumpxml_instance_data(server, hypervisor)

        cputune = dumpxml_string.findall('cputune')[0]
        emulatorpin_str = cputune.findall('emulatorpin')[0].items()[0][1]

        return emulatorpin_str

    def _check_rx_tx_from_dumpxml(self, server, hypervisor):
        """RX/TX configuration on the instance

        :param server
        :param hypervisor

        :return rx_tx_list
        """

        dumpxml_string = self._get_dumpxml_instance_data(server, hypervisor)

        devices = dumpxml_string.findall('devices')[0]
        rx_tx_list = []
        for value in devices.findall('interface')[0].findall(
                'driver')[0].items():
            rx_tx_list.append(value[1])

        return ','.join(rx_tx_list)

    def locate_dedicated_and_shared_cpu_set(self, node=None, keys=None):
        """Locate dedicated and shared cpu set

        The method locates the cpus provided by the compute for the instances.
        The cpus divided into two groups: dedicated and shared

        :param node: The node that the query should executed on.
        :param keys: The hiera mapping that should be queried.
        :return Two lists of dedicated and shared cpus set
        """
        if not node:
            hyper_kwargs = {'shell':
                            CONF.nfv_plugin_options.undercloud_rc_file}
            node = self._get_hypervisor_ip_from_undercloud(**hyper_kwargs)[0]
        if not keys:
            hiera_dedicated_cpus = "nova::compute::cpu_dedicated_set"
            hiera_shared_cpus = "nova::compute::cpu_shared_set"
            keys = [hiera_dedicated_cpus, hiera_shared_cpus]
        dedicated, shared = shell_utils.\
            retrieve_content_from_hiera(node=node, keys=keys)
        dedicated = dedicated.strip('[""]')
        dedicated = shell_utils.parse_int_ranges_from_number_string(dedicated)
        shared = shared.strip('[]').split(', ')
        shared = [int(vcpu) for vcpu in shared]
        return dedicated, shared

    def locate_numa_aware_networks(self, numa_physnets):
        """Locate numa aware networks

        :param numa_physnets: Dict of numa aware and non aware physnets
        :return numa_aware_net aware and non aware dict
        """
        numa_aware_net = self.networks_client.list_networks(
            **{'provider:physical_network':
                numa_physnets['numa_aware_net']['net'],
               'router:external': False})['networks']
        if numa_aware_net:
            numa_aware_net = numa_aware_net[0]['id']
        else:
            nets = self.networks_client.list_networks(
                **{'router:external': False})['networks']
            for net in nets:
                if net['provider:network_type'] == 'vxlan':
                    numa_aware_net = net['id']
        return numa_aware_net

    def compare_emulatorpin_to_overcloud_config(self, server, overcloud_node,
                                                config_path, check_section,
                                                check_value):
        """Compare emulatorpin to overcloud config

        :param server
        :param overcloud_node
        :param config_path
        :param check_section
        :param check_value
        """

        instance_emulatorpin = \
            self._check_emulatorpin_from_dumpxml(server, overcloud_node)
        nova_emulatorpin = shell_utils.\
            get_value_from_ini_config(overcloud_node,
                                      config_path,
                                      check_section,
                                      check_value)
        # Construct a list of integers of instance emulatorpin threads
        parsed_instance_emulatorpin = shell_utils.\
            parse_int_ranges_from_number_string(instance_emulatorpin)

        # Construct a list of integers of nova emulator pin threads
        parsed_nova_emulatorpin = shell_utils.\
            parse_int_ranges_from_number_string(nova_emulatorpin)

        # Check if all parsed instance emulatorpin threads are part of
        # configured nova emulator pin threads
        if set(parsed_instance_emulatorpin).issubset(parsed_nova_emulatorpin):
            return True
        return False

    def compare_rx_tx_to_overcloud_config(self, server, overcloud_node,
                                          config_path, check_section,
                                          check_value):
        """Compare RX/TX to overcloud config

        :param server
        :param overcloud_node
        :param config_path
        :param check_section
        :param check_value
        """

        instance_rx_tx = self._check_rx_tx_from_dumpxml(server, overcloud_node)
        nova_rx_tx = shell_utils.\
            get_value_from_ini_config(overcloud_node,
                                      config_path,
                                      check_section,
                                      check_value)

        if instance_rx_tx == nova_rx_tx:
            return True
        return False

    def check_number_queues(self):
        """This method checks the number of max queues"""
        self.ip_address = self._get_hypervisor_ip_from_undercloud(
            **{'shell': CONF.nfv_plugin_options.undercloud_rc_file})
        ovs_process_pid = shell_utils.check_pid_ovs(self.ip_address[0])
        count_pmd = "ps -T -p {} | grep pmd | wc -l".format(ovs_process_pid)
        numpmds = int(shell_utils.run_command_over_ssh(self.ip_address[0],
                                                       count_pmd))
        # We ensure that a number is being parsed, otherwise we fail
        cmd = r'sudo ovs-vsctl show | sed -n "s/.*n_rxq=.\([1-9]\).*/\\1/p"'
        numqueues = (shell_utils.
                     run_command_over_ssh(self.ip_address[0],
                                          cmd)).encode('ascii', 'ignore')
        if not isinstance(numqueues, type(str())):
            numqueues = numqueues.decode("utf-8")
        msg = "There are no queues available"
        self.assertNotEqual((numqueues.rstrip("\n")), '', msg)
        # Different multiple queues is not a supported scenario as per now
        self.assertTrue(str.isdigit(numqueues.split("\n")[0]),
                        "Queue recieved is not a digit")
        numqueues = int(numqueues.split("\n")[0])
        maxqueues = numqueues * numpmds
        return maxqueues

    def _prepare_cloudinit_file(self, install_packages=None):
        """This method creates cloud-init file with instance boot config.

        Set params:
        User credentials: user:passwd
        Enable direct (console) root login
        Set default route, add additional interface and restart network
        Configures repository
        :param install_packages: Provide the packages that should be installed.
                         Multiple packages should be separated by comma -
                         iperf,htop,vim
        """
        if not self.user_data:
            self.user_data = '''
                             #cloud-config
                             user: {user}
                             password: {passwd}
                             chpasswd: {{expire: False}}
                             ssh_pwauth: True
                             disable_root: 0
                             '''.format(user=self.instance_user,
                                        passwd=self.instance_pass)
        if (self.test_instance_repo and 'name' in
                self.test_instance_repo):
            repo_name = self.external_config['test_instance_repo']['name']
            repo_url = self.external_config['test_instance_repo']['url']
            repo = '''
                             yum_repos:
                                 {repo_name}:
                                     name: {repo_name}
                                     baseurl: {repo_url}
                                     enabled: true
                                     gpgcheck: false
                    '''.format(repo_name=repo_name,
                               repo_url=repo_url)
            self.user_data = "".join((self.user_data, repo))

        if install_packages is not None:
            header = '''
                             packages:'''
            body = ''
            for package in install_packages:
                body += '''
                             - {package}'''.format(package=package)
            package = "".join((header, body))
            self.user_data = "".join((self.user_data, package))

        # Use cloud-config write_files module to copy files
        if CONF.nfv_plugin_options.transfer_files_src and \
                CONF.nfv_plugin_options.transfer_files_dest:
            LOG.info('Locate tests scripts directory')
            exec_dir = os.path.dirname(os.path.realpath(__file__))
            scripts_dir = os.path.join(
                exec_dir, CONF.nfv_plugin_options.transfer_files_src)
            test_scripts = os.listdir(scripts_dir)
            test_scripts = [fil for fil in test_scripts if fil.endswith('.py')]

            header = '''
                             write_files:'''
            body = ''
            for file_content in test_scripts:
                file_dest = os.path.join(
                    CONF.nfv_plugin_options.transfer_files_dest, file_content)
                # The "custom_net_config" script should be placed in a
                # separate location to be executed on every boot.
                if file_content == 'custom_net_config.py':
                    file_dest = '/var/lib/cloud/scripts/per-boot/' \
                                'custom_net_config.py'
                with open(os.path.join(scripts_dir, file_content), 'r') as f:
                    content = f.read().encode('utf8')
                    content = str(base64.b64encode(content).decode('ascii'))
                    body += '''
                               - path: {file_dest}
                                 owner: root:root
                                 permissions: 0755
                                 encoding: base64
                                 content: |
                                     {file_content}
                            '''.format(file_dest=file_dest,
                                       file_content=content)
            files = "".join((header, body))
            self.user_data = "".join((self.user_data, files))

        user_data = textwrap.dedent(self.user_data).lstrip().encode('utf8')
        self.user_data_b64 = base64.b64encode(user_data)
        return self.user_data_b64

    def copy_file_to_remote_host(self, host, ssh_key, username=None,
                                 files=None, src_path=None, dst_path=None,
                                 timeout=60):
        """The method copy provided file to a specified remote host.

        Note! - The method is temporary. Should be removed once config_drive is
        implemented.

        :param host: Remote host to copy files to
        :param username: Username for the remote  host
        :param ssh_key: SSH key for the remote host
        :param files: File or comma separated file to copy
        :param src_path: Source path of the files
        :param dst_path: Destination path of the files
        :param timeout: A timeout for SSH connection to become active
        :return Return local and remote path
        """
        result = None
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        private_key_file = StringIO()
        private_key_file.write(ssh_key)
        private_key_file.seek(0)
        ssh_key = paramiko.RSAKey.from_private_key(private_key_file)

        if username is None:
            username = self.instance_user

        timeout_start = time.time()
        ssh_success = False
        while time.time() < timeout_start + timeout:
            time.sleep(2)
            try:
                ssh.connect(host, username=username, pkey=ssh_key)
                ssh_success = True
                break
            except paramiko.ssh_exception.NoValidConnectionsError:
                print('SSH transport is not ready...')
                continue
        if not ssh_success:
            raise lib_exc.TimeoutException('Instance ssh connection timed out')

        try:
            if not all([files, src_path, dst_path]):
                raise NameError('The following variables must be provided '
                                '- files, src_path, dst_path.')
        except NameError:
            raise

        sftp = ssh.open_sftp()
        for copy_file in files.split(','):
            path = os.path.dirname(__file__)
            src_path = os.path.join(path, src_path)
            file_local = src_path + '/' + copy_file
            file_remote = dst_path + '/' + copy_file

            sftp.put(file_local, file_remote)
            result = 'Copied ' + file_local + ' to ' + host + ':' + file_remote

        sftp.close()
        ssh.close()
        return result

    def ping_via_network_namespace(self, ping_to_ip, network_id):
        cmd = ("sudo ip netns exec qdhcp-" + network_id
               + " ping -c 10 " + ping_to_ip)
        ctrl_ip = urlparse(CONF.identity.uri).netloc.split(':')[0]
        result = shell_utils.run_command_over_ssh(ctrl_ip, cmd)
        for line in result.split('\n'):
            if 'packets transmitted' in line:
                LOG.info("Ping via namespace result: %s", line)
                received_str = line.split(',')[1].strip()
                try:
                    received = int(received_str.split(' ')[0])
                except ValueError:
                    break
                if received > 0:
                    return True
                break
        return False

    def _read_and_validate_external_resources_data_file(self):
        """Validate yaml file contains externally created resources"""
        LOG.info("Found external resources file. Validating...")
        with open(CONF.nfv_plugin_options.external_resources_output_file,
                  'r') as f:
            try:
                self.external_resources_data = yaml.safe_load(f)
            except yaml.YAMLError as exc:
                pm = exc.problem_mark
                raise Exception('The {} file has an issue on line {} at '
                                'position {}.'.format(pm.name,
                                                      pm.line,
                                                      pm.column))

        if self.external_resources_data['key_pair'] is None or not \
                os.path.exists(self.external_resources_data['key_pair']):
            raise Exception('The private key is missing from the yaml file.')
        for srv in self.external_resources_data['servers']:
            if not set(srv.keys()) >= {'name', 'id', 'fip', 'groups'}:
                raise ValueError('The yaml file missing of the following keys:'
                                 ' name, id or fip.')

    def _organize_external_created_resources(self, group=None):
        """Organize the external created resource by test groups"""
        groups = {}
        bulk_servers = self.external_resources_data['servers']
        for srv in bulk_servers:
            for grp in srv['groups']:
                if grp in groups:
                    groups[grp].append(srv)
                else:
                    groups[grp] = [srv]
        if group not in groups:
            raise ValueError('The required group - "{}" is missing on '
                             'existing resources'.format(group))
        servers = groups[group]

        with open(self.external_resources_data['key_pair'], 'r') as key:
            key_pair = {'private_key': key.read()}
        return servers, key_pair

    def get_ovs_interface_statistics(self, interfaces, previous_stats=None,
                                     hypervisor=None):
        """This method get ovs interface statistics

        :param interfaces: interfaces in which statistics will be retrieved
        :param previous_stats: get the difference between current stats and
                               previous stats
        :param hypervisor: hypervisor ip, if None it will be selected the first
                           one
        :return statistics
        """
        self.ip_address = self._get_hypervisor_ip_from_undercloud(
            **{'shell': CONF.nfv_plugin_options.undercloud_rc_file})
        hypervisor_ip = self.ip_address[0]
        if hypervisor is not None:
            if hypervisor not in self.ip_address:
                raise ValueError('invalid hypervisor ip {}, not in {}'
                                 .format(hypervisor,
                                         ' '.join(self.ip_address)))
            else:
                hypervisor_ip = hypervisor

        shell_utils.check_pid_ovs(hypervisor_ip)
        # We ensure that a number is being parsed, otherwise we fail
        statistics = {}
        for interface in interfaces:
            command = 'sudo ovs-vsctl get Interface {} ' \
                      'statistics'.format(interface)
            statistics[interface] = \
                yaml.safe_load(
                    shell_utils.run_command_over_ssh(
                        hypervisor_ip, command).replace(
                        '"', '').replace(
                        '{', '{"').replace(', ', ', "').replace('=', '":'))
            if previous_stats is not None and \
               interface in previous_stats.keys():
                for stat in statistics[interface].keys():
                    if stat in previous_stats[interface].keys():
                        statistics[interface][stat] -= \
                            previous_stats[interface][stat]
                    else:
                        raise ValueError('missing ovs interface stat {} '
                                         'to compare'.format(stat))

        return statistics

    def get_ovs_multicast_groups(self, switch, multicast_ip=None,
                                 hypervisor=None):
        """This method get ovs multicast groups

        :param switch: ovs switch to get multicast groups
        :param multicast_ip: filter by multicast ip
        :param hypervisor: hypervisor ip, if None it will be selected the first
                           one
        :return multicast groups
        """
        self.ip_address = self._get_hypervisor_ip_from_undercloud(
            **{'shell': CONF.nfv_plugin_options.undercloud_rc_file})
        hypervisor_ip = self.ip_address[0]
        if hypervisor is not None:
            if hypervisor not in self.ip_address:
                raise ValueError('invalid hypervisor ip {}, not in {}'
                                 .format(hypervisor,
                                         ' '.join(self.ip_address)))
            else:
                hypervisor_ip = hypervisor

        shell_utils.check_pid_ovs(hypervisor_ip)

        command = 'sudo ovs-appctl mdb/show {}'.format(switch)
        output = list(filter(None,
                             shell_utils.run_command_over_ssh(
                                 hypervisor_ip, command).split('\n')))
        fields = None
        output_data = []
        for line in output:
            data = list(filter(None, line.split(" ")))
            if fields is None:
                fields = data
            else:
                data = dict(zip(fields, data))
                if multicast_ip is None or \
                   multicast_ip is not None and data['GROUP'] == multicast_ip:
                    output_data.append(data)
        return output_data

    def _get_hypervisor_host_ip(self, name=None):
        """Get hypervisor ip

        This Method lists aggregation based on name,
        and returns the aggregated search for IP through Hypervisor list API
        Add support in case of NoAggregation, and Hypervisor list is not empty
        if host=None, no aggregation, or name=None and if hypervisor list has
        one member return the member

        :param name
        """
        host = None
        ip_address = ''
        if name:
            host = self._list_aggregate(name)

        hyper = self.manager.hypervisor_client.list_hypervisors()
        if host:
            for i in hyper['hypervisors']:
                if i['hypervisor_hostname'] == host:
                    ip_address = \
                        self.manager.hypervisor_client.show_hypervisor(
                            i['id'])['hypervisor']['host_ip']
        else:
            for i in hyper['hypervisors']:
                if i['state'] == 'up':
                    ip_address = \
                        self.manager.hypervisor_client.show_hypervisor(
                            i['id'])['hypervisor']['host_ip']
        return ip_address

    def _get_hypervisor_ip_from_undercloud(self, **kwargs):
        """This Method lists aggregation based on name

        Returns the aggregated search for IP through Hypervisor list API
        Add support in case of NoAggregation, and Hypervisor list is not empty
        if host=None, no aggregation, or name=None and if hypervisor list has
        one member return the member
        :param kwargs['shell']
        :param kwargs['server_id']
        :param kwargs['aggregation_name']
        :param kwargs['hyper_name']
        """
        host = None
        ip_address = ''
        if 'aggregation_name' in kwargs:
            host = self._list_aggregate(kwargs['aggregation_name'])

        hyper = self.manager.hypervisor_client.list_hypervisors()
        """
        if hosts in aggregations
        """
        if host:
            host_name = re.split(r"\.", host[0])[0]
            if host_name is None:
                host_name = host

            for i in hyper['hypervisors']:
                if i['hypervisor_hostname'] == host[0]:
                    command = 'openstack ' \
                              'server show ' + host_name + \
                              ' -c \'addresses\' -f value | cut -d\"=\" -f2'
                    ip_address = shell_utils.\
                        run_local_cmd_shell_with_venv(command,
                                                      kwargs['shell'])
        else:
            """
            no hosts in aggregations, select with 'server_id' in kwargs
            """
            compute = 'compute'
            if 'hyper_name' in kwargs:
                compute = kwargs['hyper_name']
            if 'server_id' in kwargs:
                server = self. \
                    os_admin.servers_client.show_server(kwargs['server_id'])
                compute = \
                    server['server']['OS-EXT-SRV-ATTR:host'].partition('.')[0]

            for i in hyper['hypervisors']:
                if i['state'] == 'up':
                    if i['hypervisor_hostname'].split(".")[0] == compute:
                        compute = i['hypervisor_hostname'].split(".")[0]
                    command = 'openstack server list -c \'Name\' -c ' \
                              '\'Networks\' -f value | grep -i {0} | ' \
                              'cut -d\"=\" -f2'.format(compute)
                    ip_address = shell_utils.\
                        run_local_cmd_shell_with_venv(command,
                                                      kwargs['shell'])

        return ip_address

    def locate_ovs_physnets(self, node=None, keys=None):
        """Locate ovs existing physnets

        The method locate the ovs existing physnets and create a dict with
        the numa aware and non aware physnets.

        :param node: The node that the query should executed on.
        :param keys: The hiera mapping that should be queried.
        :return The numa physnets dict is returned
        """
        if node is None:
            hyper_kwargs = {'shell':
                            CONF.nfv_plugin_options.undercloud_rc_file}
            node = self._get_hypervisor_ip_from_undercloud(**hyper_kwargs)[0]
        network_backend = self.dicover_deployment_network_backend(node=node)
        if not keys:
            if network_backend == 'ovs':
                hiera_bridge_mapping = \
                    "neutron::agents::ml2::ovs::bridge_mappings"
            elif network_backend == 'ovn':
                hiera_bridge_mapping = "ovn::controller::ovn_bridge_mappings"
            else:
                hiera_bridge_mapping = None
            hiera_numa_mapping = "nova::compute::neutron_physnets_numa_" \
                                 "nodes_mapping"
            hiera_numa_tun = "nova::compute::neutron_tunnel_numa_nodes"
            hiera_pci_whitelist = "nova::compute::pci::passthrough"
            keys = [hiera_bridge_mapping, hiera_numa_mapping, hiera_numa_tun,
                    hiera_pci_whitelist]
        numa_phys_content = shell_utils.retrieve_content_from_hiera(node=node,
                                                                    keys=keys)
        # Identify the numa aware physnet
        numa_aware_phys = {}
        bridge_mapping = []
        numa_aware_tun = []
        pci_whitelist = []
        for physnet in numa_phys_content:
            if 'physical_network' in physnet:
                pci_whitelist = yaml.safe_load(physnet.replace('=>', ':'))
            elif '=>' in physnet:
                numa_aware_phys = yaml.safe_load(physnet.replace('=>', ':'))
            elif ':' in physnet:
                bridge_mapping = yaml.safe_load(physnet)
            else:
                numa_aware_tun = yaml.safe_load(physnet)

        numa_physnets = {'numa_aware_net': {},
                         'non_numa_aware_net': [],
                         'numa_aware_tunnel': {}}
        physnet_list = []
        for item in bridge_mapping:
            physnet = item.split(':')[0]
            physnet_list.append(physnet)

        for physnet in physnet_list:
            if physnet in numa_aware_phys.keys():
                LOG.info('The {} is a numa aware network'.format(physnet))
                numa_physnets['numa_aware_net'] = \
                    {'net': physnet, 'numa_node': numa_aware_phys[physnet][0]}
            else:
                LOG.info('The {} is a non numa aware network'.format(physnet))
                numa_physnets['non_numa_aware_net'].append(physnet)

        # Exclude sriov networks from non numa aware list
        sriov_nets = [sriov_net['physical_network']
                      for sriov_net in pci_whitelist]
        sriov_nets = list(set(sriov_nets))
        numa_physnets['non_numa_aware_net'] = \
            [non_numa for non_numa in numa_physnets['non_numa_aware_net']
             if non_numa not in sriov_nets]

        if numa_aware_tun:
            numa_physnets['numa_aware_tunnel'] = \
                {'numa_node': numa_aware_tun[0]}
        return numa_physnets

    def list_available_resources_on_hypervisor(self, hypervisor):
        """List available CPU and RAM on dedicated hypervisor"""
        hyp_list = self.os_admin.hypervisor_client.list_hypervisors()[
            'hypervisors']
        if not any(hypervisor in a['hypervisor_hostname'] for a in hyp_list):
            raise ValueError('Specifyed hypervisor has not been found.')

        osp_release = self.get_osp_release()
        if osp_release >= 16:
            cmd = "openstack hypervisor list -c ID -c " \
                  "'Hypervisor Hostname' --format value"
            hypers = shell_utils.\
                run_local_cmd_shell_with_venv(cmd,
                                              '/home/stack/overcloudrc')
            host_id = None
            for hyper_id, hostname in zip(hypers[::2], hypers[1::2]):
                if hostname in hypervisor:
                    host_id = hyper_id
            hyper_resources = {'resources': 'VCPU:1,PCPU:1,MEMORY_MB:1'}
            hyper_info = \
                self.os_admin.placement_client.list_allocation_candidates(
                    **hyper_resources)
            hyper_info = hyper_info['provider_summaries'][host_id]['resources']
            pcpu_total = hyper_info['PCPU']['capacity']
            pcpu_used = hyper_info['PCPU']['used']
            pcpu_free = \
                hyper_info['PCPU']['capacity'] - hyper_info['PCPU']['used']
            pcpu_free_per_numa = hyper_info['PCPU']['capacity'] \
                // 2 - hyper_info['PCPU']['used']
            vcpu_total = hyper_info['VCPU']['capacity']
            vcpu_used = hyper_info['VCPU']['used']
            vcpu_free = \
                hyper_info['VCPU']['capacity'] - hyper_info['VCPU']['used']
            vcpu_free_per_numa = hyper_info['VCPU']['capacity'] \
                // 2 - hyper_info['VCPU']['used']
            ram_free = (hyper_info['MEMORY_MB']['capacity']
                        - hyper_info['MEMORY_MB']['used']) // 1024
        else:
            hyper_id = self.os_admin.hypervisor_client.search_hypervisor(
                hypervisor)['hypervisors'][0]['id']
            hyper_info = self.os_admin.hypervisor_client.show_hypervisor(
                hyper_id)['hypervisor']
            pcpu_total = hyper_info['vcpus']
            pcpu_used = hyper_info['vcpus_used']
            pcpu_free = hyper_info['vcpus'] - hyper_info['vcpus_used']
            pcpu_free_per_numa = \
                hyper_info['vcpus'] // 2 - hyper_info['vcpus_used']
            vcpu_total = None
            vcpu_used = None
            vcpu_free = None
            vcpu_free_per_numa = None
            ram_free = hyper_info['free_ram_mb'] // 1024
        return {'pcpu_total': pcpu_total, 'pcpu_used': pcpu_used,
                'pcpu_free': pcpu_free,
                'pcpu_free_per_numa': pcpu_free_per_numa,
                'vcpu_total': vcpu_total, 'vcpu_used': vcpu_used,
                'vcpu_free': vcpu_free,
                'vcpu_free_per_numa': vcpu_free_per_numa,
                'ram_free': ram_free}

    def dicover_deployment_network_backend(self, node=None):
        """Locate deployment's network backend

        The method discovers the network backend used in deployment.
        It depends on hieradata being present on the node.

        :param node: The node that the query should executed on.
        :return The deployment network backend.
        """
        # Initialize parameters
        network_backend = 'unknown'
        hieradata_keys = [
            'enabled_services'
        ]
        if node is None:
            hyper_kwargs = {'shell':
                            CONF.nfv_plugin_options.undercloud_rc_file}
            node = self._get_hypervisor_ip_from_undercloud(**hyper_kwargs)[0]
        hiera_response = \
            shell_utils.retrieve_content_from_hiera(node=node,
                                                    keys=hieradata_keys)
        # Construct a list of enabled services from response string
        enabled_services = \
            re.sub(r'\[|\]|"| ', '', hiera_response[0]).split(',')
        if 'neutron_plugin_ml2' in enabled_services:
            network_backend = 'ovs'
        elif 'neutron_plugin_ml2_ovn' in enabled_services:
            network_backend = 'ovn'
        LOG.info("Discovered network backend '{}'".format(network_backend))
        return network_backend



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from importlib import import_module
from nfv_tempest_plugin.tests.common import shell_utilities as shell_utils
from nfv_tempest_plugin.tests.scenario import baremetal_manager
from oslo_log import log as logging
from tempest import clients
from tempest.common import credentials_factory as common_creds
from tempest import config

CONF = config.CONF
LOG = logging.getLogger('{} [-] nfv_plugin_test'.format(__name__))


class BaseTest(baremetal_manager.BareMetalManager):
    def __init__(self, *args, **kwargs):
        super(BaseTest, self).__init__(*args, **kwargs)

    @classmethod
    def get_client_manager(cls, credential_type=None, roles=None,
                           force_new=None):
        """Method, manages two client managers.

        Neutron tempest plugin ,maintain its own client.manager,
        it also save upstream manager w/ primary credential.
        This methos maintain nfv_tempest_plugin upstream client.manager
        and save neutron_tempest_plugin in class member
        """
        manager = super(BaseTest, cls).get_client_manager(
            credential_type=credential_type,
            roles=roles,
            force_new=force_new
        )
        """Apply dynamic package check to import nfv_tempest_plugin service
        Load packages only when use_neutron_api_v2 is set and
        neutron_tempest_plugin is installed
        """
        if CONF.nfv_plugin_options.use_neutron_api_v2 and \
                credential_type == 'admin':
            try:
                import_module('neutron_tempest_plugin')
                network_client_v2 = \
                    import_module('nfv_tempest_plugin.services'
                                  '.network_client_v2')
                cls.os_admin_v2 = \
                    network_client_v2.Manager(manager.credentials)
            except ImportError:
                LOG.info("Failed to load neutron_tempest_plugin, \
                         please check use_neutron_api_v2 set to False")
        return manager

    @classmethod
    def setup_credentials(cls):
        """Do not create network resources for these tests

        Using public network for ssh
        """
        cls.set_network_resources()
        super(BaseTest, cls).setup_credentials()
        cls.manager = clients.Manager(
            credentials=common_creds.get_configured_admin_credentials())

    def setUp(self):
        """Set up a single tenant with an accessible server.

        If multi-host is enabled, save created server uuids.
        """
        super(BaseTest, self).setUp()
        # pre setup creations and checks read from config files

    def verify_provider_networks(self, servers=None, key_pair=None):
        """Verifies provider networks attached to guest

        This functions attempts to verify all provider networks present
        inside guest.

        If multiple servers are created, guests will attempt to ping between
        themselves on each provider network.

        :param servers: List of servers created
        :param key-pair: Key pair used to authenticate with server
        """
        for server in servers:
            # Initialize a custom key inside server object
            server['provider_networks'] = []
            # Fetch all ports assigned to server
            ports =  \
                self.os_admin.ports_client.list_ports(device_id=server['id'])
            for port in ports['ports']:
                provider_dict = {
                    'network_id': port['network_id'],
                    'mac_address': port['mac_address'],
                    'ip_address': port['fixed_ips'][0]['ip_address']
                }
                server['provider_networks'].append(provider_dict)
            # Create an SSH connection to server
            ssh_client = self.get_remote_client(server['fip'],
                                                self.instance_user,
                                                key_pair['private_key'])
            shell_utils.\
                check_guest_interface_config(ssh_client,
                                             server['provider_networks'],
                                             server['name'])

        self.check_guest_provider_networks(servers, key_pair)

    def create_and_verify_resources(self, test=None, fip=None, **kwargs):
        """Create and verify resources method

        The create and verify resources method performs basic steps in order
        to prepare the environment for the actual test.
        Create all the resources and boot an instance.
        Verify ping and SSH connection to the instance.

        The method should be used by the tests as a starting point for
        environment preparation.

        :param test: Test name from the external config file.

        :return servers, key_pair
        """
        LOG.info('Starting the {} test'.format(test))
        if fip is None:
            fip = self.fip

        servers, key_pair = self.create_server_with_resources(test=test,
                                                              fip=fip,
                                                              **kwargs)

        for srv in servers:
            LOG.info('Instance details: fip: {}, instance_id: {}'.format(
                srv['fip'], srv['id']))

            srv['hypervisor_ip'] = self._get_hypervisor_ip_from_undercloud(
                **{'shell': CONF.nfv_plugin_options.undercloud_rc_file,
                   'server_id': srv['id']})[0]
            self.assertNotEmpty(srv['hypervisor_ip'],
                                "_get_hypervisor_ip_from_undercloud "
                                "returned empty ip list")

            LOG.info('Test {} instance connectivity.'.format(srv['fip']))
            if fip:
                self.check_instance_connectivity(ip_addr=srv['fip'],
                                                 user=self.instance_user,
                                                 key_pair=key_pair[
                                                     'private_key'])
            else:
                LOG.info("FIP is disabled, ping %s using network namespaces" %
                         srv['fip'])
                ping = self.ping_via_network_namespace(srv['fip'],
                                                       srv['network_id'])
                self.assertTrue(ping)

        # Verify provider networks only when requested and if FIP is assigned
        if self.test_all_provider_networks and fip:
            self.verify_provider_networks(servers, key_pair)

        return servers, key_pair



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

nfv_plugin_options = cfg.OptGroup(name="nfv_plugin_options",
                                  title="NFV plugin params")

NfvPluginOptions = [
    cfg.StrOpt('overcloud_node_user',
               default='heat-admin',
               help="SSH user for overcloud node - controller/compute"),
    cfg.StrOpt('overcloud_node_pass',
               default=None,
               help="SSH password for overcloud node - controller/compute"),
    cfg.StrOpt('overcloud_node_pkey_file',
               default='/home/stack/.ssh/id_rsa',
               help="SSH private key path for overcloud node - "
                    "controller/compute"),
    cfg.StrOpt("instance_user",
               default="cloud-user",
               help="SSH user for the guest instance"),
    cfg.StrOpt("instance_pass",
               default="password",
               help="SSH password for the guest instance"),
    cfg.StrOpt('external_config_file',
               default=None,
               help="The path to yml file for additional configurations"),
    cfg.StrOpt('transfer_files_src',
               default='tests_scripts',
               help="Relative directory path that contains the test scripts"),
    cfg.StrOpt('transfer_files_dest',
               default='/usr/local/bin/nfv_scripts/',
               help="The default location of nfv test scripts"),
    cfg.StrOpt('external_resources_output_file',
               default='/home/stack/resources_output_file.yml',
               help="The path to the file output of the created resources"),
    cfg.StrOpt('quota_cores',
               default=40,
               help="The numbers of cpu cores for the tenant use"),
    cfg.StrOpt('quota_ram',
               default=81920,
               help="The numbers of ram for the tenant use"),
    cfg.BoolOpt('test_all_provider_networks',
                default=False,
                help="Verify provider networks attached to guests"),
    cfg.StrOpt('login_security_group_rules',
               default='[{"protocol": "tcp", "direction": "ingress"'
                       ','
                       '"port_range_max": "22", "port_range_min": "22"}'
                       ','
                       '{"protocol": "icmp", "direction": "ingress"}]',
               help='Configuration for test security groups.'
                    'The format is JSON. '
                    '"protocol":<icmp/udp/tcp> - string '
                    '"direction":<ingress/egress> - string '
                    '"port_range_max":<max_port_range> - string '
                    '"port_range_min":<min_port_range> - string '),
    cfg.StrOpt('max_qos_rules',
               default='[{"max_kbps": 4000000, "max_burst_kbps": 4000000 }'
                       ','
                       ' {"max_kbps": 9000000, "max_burst_kbps": 4000000 }]',
               help='Configuration for max qos rules.'
                    'The format is JSON. '
                    '"max_kbps":<kbps> - string '
                    '"max_burst_kbps":<kbps> - string '),
    cfg.StrOpt('min_qos_rules',
               default='[{"min_kbps": 25000000}]',
               help='Configuration for min qos rules.'
                    'The format is JSON. '
                    '"min_kbps":<kbps> - string '),
    cfg.StrOpt('undercloud_rc_file',
               default='/home/stack/stackrc',
               help="Full path to undercloud rc file"
                    "usually called stackrc"),
    cfg.BoolOpt('use_neutron_api_v2',
                default=False,
                help="Use neutron-tempest-plugin clients"),
    cfg.IntOpt('hypervisor_wait_timeout',
               default=300,
               help='Timeout in seconds to wait for the '
                    'hypervisor reachability'),

]



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nfv_tempest_plugin.tests.common import shell_utilities as shell_utils
from nfv_tempest_plugin.tests.scenario import base_test
from oslo_log import log as logging
from tempest import config

CONF = config.CONF
LOG = logging.getLogger('{} [-] nfv_plugin_test'.format(__name__))


class TestNfvOffload(base_test.BaseTest):
    def __init__(self, *args, **kwargs):
        super(TestNfvOffload, self).__init__(*args, **kwargs)
        self.hypervisor_ip = None

    def setUp(self):
        """Set up a single tenant with an accessible server.

        If multi-host is enabled, save created server uuids.
        """
        super(TestNfvOffload, self).setUp()

    def test_offload_ovs_config(self):
        """Check ovs config for offload on all hypervisors

        """
        # Command to check if hw-offload is enabled in OVS
        cmd = ("sudo ovs-vsctl get open_vswitch . "
               "other_config:hw-offload")
        # Retrieve all hypvervisors
        hypervisors = self._get_hypervisor_ip_from_undercloud(
            shell=CONF.nfv_plugin_options.undercloud_rc_file)
        # Intialize results list
        result = []
        # Expected result is a list of dicts, each dict contains
        # a key which is hypervisor's IP and the value 'true'
        # Example:
        # [{192.0.60.1: 'true'}, {192.0.60.2: 'true'}]
        expected_result = [{ip: 'true'} for ip in hypervisors]
        for hypervisor in hypervisors:
            out = shell_utils.run_command_over_ssh(hypervisor, cmd)
            if out:
                # Strip newlines and remove double quotes
                output = out.rstrip().replace('"', '')
            # HW-Offload not enabled if no text returned
            else:
                output = 'false'
            LOG.info("Hypervisor '{h}' is OVS HW-offload "
                     "capable: '{r}'".format(h=hypervisor,
                                             r=output))
            result.append({hypervisor: output})
        msg = "Not all hypervisors have OVS HW-Offload enabled"
        self.assertItemsEqual(expected_result, result, msg)

    def test_offload_nic_eswitch_mode(self, test='offload'):
        """Check eswitch mode of nic for offload on all hypervisors

        :param test: Test name from the external config file.
        """
        test_dict = self.test_setup_dict[test]
        if 'offload_nics' in test_dict:
            offload_nics = test_dict['offload_nics']
        else:
            raise ValueError('offload_nics is not defined in offload test')
        # Retrieve all hypvervisors
        hypervisors = self._get_hypervisor_ip_from_undercloud(
            shell=CONF.nfv_plugin_options.undercloud_rc_file)
        # ethtool cmd to retrieve PCI bus of interface
        ethtool_cmd = ("sudo ethtool -i {} | grep bus-info "
                       "| cut -d ':' -f 2,3,4 | awk '{{$1=$1}};1'")
        # devlink cmd to retrieve switch mode of interface
        devlink_cmd = "sudo devlink dev eswitch show pci/{}"
        # Intialize results list
        result = []
        # Expected result is a list of dicts containing a dict of
        # hypervisor's IP, its offload nics as keys and the value 'true'
        # Example:
        # [{'192.0.160.1': [{'p6p1': 'true'}, {'p6p2': 'true'}]},
        #  {'192.0.160.2': [{'p6p1': 'true'}, {'p6p2': 'true'}]}]
        expected_result = [{ip: [{nic: 'true'} for nic in offload_nics]}
                           for ip in hypervisors]
        for hypervisor in hypervisors:
            dev_result = []
            for nic in offload_nics:
                pci = shell_utils.run_command_over_ssh(hypervisor,
                                                       ethtool_cmd.format(nic))
                dev_query = shell_utils.\
                    run_command_over_ssh(hypervisor,
                                         devlink_cmd.format(pci))
                if 'switchdev' in dev_query:
                    output = 'true'
                else:
                    output = 'false'
                LOG.info("Hypervisor '{h}' NIC '{n}' is in switchdev mode: {r}"
                         .format(h=hypervisor, n=nic, r=output))
                dev_result.append({nic: output})
            result.append({hypervisor: dev_result})
        msg = "Not all hypervisors contains nics in switchev mode"
        self.assertItemsEqual(expected_result, result, msg)

    def test_offload_ovs_flows(self, test='offload_flows'):
        """Check OVS offloaded flows

        The following test deploy vms, on hw-offload computes.
        It sends async ping and check offload flows exist in ovs.

        :param test: Test name from the external config file.
        """

        LOG.info('Start test_offload_ovs_flows test.')
        LOG.info('test_offload_ovs_flows create vms')
        # Create servers
        servers, key_pair = self.create_and_verify_resources(test=test,
                                                             num_servers=4)
        cmd = 'sudo ovs-appctl dpctl/dump-flows type=offloaded'
        # Iterate over created servers
        for server in servers:

            shell_utils.continuous_ping(server['fip'],
                                        duration=30)
            LOG.info('test_offload_ovs_flows verify flows on geust {}'.
                     format(server['fip']))

            out = shell_utils.\
                run_command_over_ssh(server['hypervisor_ip'],
                                     cmd)
            ports =  \
                self.os_admin.ports_client.list_ports(device_id=server['id'])
            msg = ('Port with mac address {} is expected to be part of '
                   'offloaded flows')
            for port in ports['ports']:
                if 'capabilities' in port['binding:profile'] and 'switchdev'\
                        in port['binding:profile']['capabilities']:
                    self.assertIn(port['mac_address'], out,
                                  msg.format(port['mac_address']))
        # Pings are running check flows exist
        # Retrieve all hypvervisors
        hypervisors = self._get_hypervisor_ip_from_undercloud(
            shell=CONF.nfv_plugin_options.undercloud_rc_file)
        # Command to check offloaded flows in OVS
        cmd = 'sudo ovs-appctl dpctl/dump-flows type=offloaded'
        for hypervisor in hypervisors:
            out = shell_utils.run_command_over_ssh(hypervisor,
                                                   cmd)
            msg = 'Hypervisor {} has no offloaded flows in OVS'.format(
                hypervisor)
            self.assertNotEmpty(out, msg)
            LOG.info('Hypercisor {} has offloaded flows in OVS'.format(
                hypervisor))

        # send stop statistics signal
        shell_utils.stop_continuous_ping()
